/*************************************************************************\
 * File   : TkImageLoad.c
 * $Id: tkImageLoad.c,v 1.3 2007/04/05 10:27:28 harry Exp $ 
 * Purpose: Load xray diffraction image files as tk photo images
 *
 * Author : Geoff Battye
 * Dated  : 10/07/2006
 *
\************************************************************************/

#include <stdio.h>
#include <tcl.h>
#include <tk.h>
#include <assert.h>

#define HEADER_LINE_LENGTH 512
#define BYTES_PER_PIXEL 2
#define BLOCKSIZE 20
//#define BLOCKED 1
//#define UNBLOCKED 1
#define IN_BLOCK_FLIP 1

/* *********************************************************************/

/* readHeaderAdsc - read the header of an ADSC image file to get the
 *                  image data offset (bytes) and image height and width
 *
 * Parameters: filename - name of image file
 *             offset   - place to store number of bytes before image data
 *             width    - place to store image width
 *             height   - place to store image height
 */


int readHeaderAdsc(char *filename, int *offset, int *width, int *height) {
  FILE *f ;
  char line[HEADER_LINE_LENGTH] ;
  char number[12] ;
  int end_of_header= 0 ;
  

  f= fopen(filename,"r") ;
  
  while (!end_of_header) {
    fgets(line, HEADER_LINE_LENGTH, f) ;    
    if (strncmp(line,"HEADER_BYTES",12) == 0) {
      strncpy(number,&line[13],5) ;
      *offset= atoi(number) ;
      strncpy(number,"\0",12) ;
    } else {
      if (strncmp(line,"SIZE",4) == 0) {
	if (line[4] == '1') {
	  strncpy(number,&line[6],4) ;
	  *width= atoi(number) ;
	  strncpy(number,"\0",12) ;
	} else {
	  if (line[4] == '2') {
	    strncpy(number,&line[6],4) ;
	    *height= atoi(number) ;
	    strncpy(number,"\0",12) ;
	  }
	}
      } else {
	if (strncmp(line,"}",1) == 0) {
	  end_of_header= 1 ;
	}
      }
    }
  }
  return(0) ;
}

/* readData - read raw data from a file
 *
 * Parameters: data   - pointer to address of place to put the data
 *             infile - filename
 *             offset - number of bytes in file before image data
 *             nbytes - the number of bytes of image datat to be read
 */

int readData(unsigned short int **data, char *infile, int offset, int nbytes) {
  FILE *f ;
  size_t bytes_read ; 
  int i ;

  unsigned short int *data_pointer ;

  f= fopen(infile,"rb") ;
  *data= (unsigned short int*) Tcl_AttemptAlloc(nbytes+offset);
  data_pointer= *data ;

  /*data = malloc(nbytes+offset);*/
  bytes_read = fread(data_pointer,1,nbytes+offset,f);
  
  return(bytes_read);
}

/* scaleData - scale image data from two-byte raw data to single byte
 *             scaled data
 *
 * Parameters: scaled_data   - address of place to put the scaled data
 *             unscaled_data - location of the unscaled data
 *             width, height - image dimensions
 *             min, max      - raw values limits to be scaled to 0 and 255
 */

int scaleData(unsigned char **scaled_data, unsigned short int *unscaled_data, int width, int height, int *min, int *max, int *pixel_histogram) {

  float range ;
  register unsigned int i ;
  double total ;
  unsigned char lookup255[32768] ;
  unsigned char *scaled_data_pointer ;

  /* Allocate memory for scaled image data */
  *scaled_data = (unsigned char*) Tcl_AttemptAlloc(width*height);
  
  scaled_data_pointer= *scaled_data ;
  
  /* Calcluate min and max intensities, if not provided */
  if ((*max == 0) && (*min == 0)) {
    /* we need to manually determine the limits on the pixel values -
       this should be done over the image as a whole */

    total = 0;
    for(i = 0; i < (width*height) ; ++i) {
      total+= unscaled_data[i] ;
    }
    *min= 0 ;
    *max= 4.0 * total / (width*height);
  } 
  
  /* make a lookup table for converting raw data to 0-255 greyscale */
  range= 255.0 / (*max - *min) ;
  for (i= 0 ; i <= 32767 ; ++i) {
    if (i <= *min) {
      lookup255[i]= 255 ;
    } else if (i >= *max) {
      lookup255[i]= 0 ;
    } else {
      lookup255[i]= (int) (*max-i)*range;
    }
  }

  /* build image */
  for (i= 0 ; i < (width*height) ; ++i) {
    if (unscaled_data[i] > 32767) {
      unscaled_data[i]= 32767 ;
    }
    scaled_data_pointer[i]= lookup255[unscaled_data[i]] ;
    pixel_histogram[unscaled_data[i]>>6]++ ;
  }

  return(0) ;
}

/* createTkImage - put the scaled data into a tk photo image block
 *
 * Parameters: interp - tcl interpreter
 *             data   - scaled image data
 *             width  - image width
 *             height - image height
 *             dstBlockPtr - pointer to destination image block
 */

int createTkImage (
		   Tcl_Interp *interp,
		   unsigned char *newimage,
		   int width,
		   int height,
		   Tk_PhotoImageBlock *dstBlockPtr) {
    
  dstBlockPtr->width = width;
  dstBlockPtr->height = height;
  dstBlockPtr->pixelSize = 1;
  dstBlockPtr->pitch = width;
  dstBlockPtr->pixelPtr = newimage;
  dstBlockPtr->offset[0] = 0;
  dstBlockPtr->offset[1] = 0;
  dstBlockPtr->offset[2] = 0;
  dstBlockPtr->offset[3] = 0;

   return TCL_OK;
}

/* Declarations of functions to be created as Tk commands */

static int imageLoad  (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
static int imageTranspose (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) ;

/* createTkCommands - create Tk commands from C functions */

int createTkCommands(Tcl_Interp *interp)
{
   Tcl_CreateObjCommand(interp, "imageLoad",   imageLoad,   (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
   Tcl_CreateObjCommand(interp, "imageTranspose",   imageTranspose,   (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
   return 1;
}

/* imageLoad - loads data from a image file into a Tk Photo image
 *
 * Parameters: specified by Tcl command object interface
 */

static int imageLoad (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) {

  if (objc != 6) {
    Tcl_WrongNumArgs(interp, 1, objv, "filetype filename destImg min max");
    return TCL_ERROR;
  }

  int result;
  char *filetype, *filename, *dest;
  Tk_PhotoHandle dp;
  Tk_PhotoImageBlock db;
  Tcl_Obj *listPtr ;
  int offset, width, height, min, max ;
  unsigned short int *data ;
  unsigned char *newimage ;
  register int i ;
  int pixel_histogram[512] ;

  /* Get arguments */
  filetype = Tcl_GetString(objv[1]);
  filename = Tcl_GetString(objv[2]);
  dest = Tcl_GetString(objv[3]);
  if (Tcl_GetIntFromObj(interp,objv[4],&min) != TCL_OK) {
    return TCL_ERROR ;
  }
  if (Tcl_GetIntFromObj(interp,objv[5],&max) != TCL_OK) {
    return TCL_ERROR ;
  }

  /* Initialize array used to store pixel intensity histogram */
  for (i= 0 ; i < 512 ; i++) {
    pixel_histogram[i]= 0 ;
  }

  /* Get image */
  dp = Tk_FindPhoto(interp, dest);
  if (dp==0) {
    Tcl_AppendResult (interp, "Invalid image destination specified",NULL);
    return TCL_ERROR;
  }
  db.pixelPtr = 0;
  Tk_PhotoGetImage(dp, &db);

  if (strcmp("ADSC",filetype) == 0) {
    /* read header to get data offset and image size */
    readHeaderAdsc(filename, &offset, &width, &height) ;
  } else {
    Tcl_AppendResult (interp, "Invalid file type specified",NULL);
    return TCL_ERROR;
  }
    
  /* Read image data from file */
  readData(&data,filename,offset,width*height*BYTES_PER_PIXEL) ;

  /* Scale data putting result in newimage */
  scaleData(&newimage,&data[offset/BYTES_PER_PIXEL],width,height,&min,&max,pixel_histogram) ;

  /* Free memory holding raw data */
  if (data) {
    Tcl_Free((char *)data);
  }

  /* Format data into TkImage */
  result = createTkImage(interp,newimage,width,height,&db);
  if (result != TCL_OK) {
    return result ;
  }
  /* put the photo block into the image */
#ifdef __alpha
  Tk_PhotoPutBlock(dp, &db, 0, 0, width, height);
#else
  Tk_PhotoPutBlock(dp, &db, 0, 0, width, height, TK_PHOTO_COMPOSITE_SET);
#endif

  /* Free the photo block's memory now it is finished with */
  if (db.pixelPtr) {
    Tcl_Free((char *)db.pixelPtr);
  }

  /* Create a Tcl list object with min, max and histogram */
  listPtr = Tcl_NewListObj( 0, NULL);
  for (i= 0 ; i < 512 ; ++i) {
    Tcl_ListObjAppendElement(interp,listPtr,Tcl_NewIntObj(pixel_histogram[i])) ;
  }
  Tcl_ListObjAppendElement(interp,listPtr,Tcl_NewIntObj(min)) ;
  Tcl_ListObjAppendElement(interp,listPtr,Tcl_NewIntObj(max)) ;

  /* Set the Tcl list as the command's returned object */
  Tcl_SetObjResult(interp,listPtr) ;

  return TCL_OK;
}


/* imageTranspose - transposes a Tk Photo image
 *
 * Parameters: specified by Tcl command object interface
 */

static int imageTranspose (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) {

  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Image");
    return TCL_ERROR;
  }

  int result;
  char *image ;
  Tk_PhotoHandle photo;
  Tk_PhotoImageBlock block;
  unsigned char *old_image ;
  unsigned char *new_image ;
  register int width ;
  register int height ;
  register int i ;
#ifndef IN_BLOCK_FLIP
  register int j ;
  register int k ;
#endif //IN_BLOCK_FLIP
#ifdef BLOCKED
  register int x ;
  register int y ;
  register int block_width ;
  register int block_height ;
#endif //BLOCKED

  /* Get arguments */
  image = Tcl_GetString(objv[1]);

  /* Get image */
  photo = Tk_FindPhoto(interp, image);
  if (photo==0) {
    Tcl_AppendResult (interp, "Invalid image specified",NULL);
    return TCL_ERROR;
  }
  block.pixelPtr = 0;
  Tk_PhotoGetImage(photo, &block);
    
  width= block.width ;
  height= block.height ;
  //  printf("Transposing %ix%i image,\n",width,height) ;

  old_image= block.pixelPtr ;

#ifndef IN_BLOCK_FLIP
  new_image= (unsigned char*)Tcl_AttemptAlloc(block.width*block.height*block.pixelSize);
#else // !IN_BLOCK_FLIP
  new_image= (unsigned char*)Tcl_AttemptAlloc(block.width*block.height);
#endif // IN_BLOCK_FLIP

#ifdef UNBLOCKED
  for (i= 0 ; i < width ; ++i) {
    for (j= 0 ; j < height ; ++j) {
      for (k= 0 ; k < 4 ; ++k) {
	new_image[(i*height*4)+(j*4)+k]= old_image[(i*4)+(j*width*4)+k] ;
      }
    }
  }
  block.width= height ;
  block.height= width ;
  block.pitch= block.pixelSize * block.width ;
  block.pixelPtr= new_image ;
#else // !UNBLOCKED
#ifdef BLOCKED
  for (x= 0 ; x <= (width / BLOCKSIZE) ; ++x) {
    if (x == (width / BLOCKSIZE)) {
      block_width= (width % BLOCKSIZE) ;
    } else {
      block_width= BLOCKSIZE ;
    }
    for (y= 0 ; y <= (height / BLOCKSIZE) ; ++y) {
      if (y == (height / BLOCKSIZE)) {
	block_height= (height % BLOCKSIZE) ;
      } else {
	block_height= BLOCKSIZE ;
      }
      for (i= 0 ; i < block_width ; ++i) {
	for (j= 0 ; j < block_height ; ++j) {
	  for (k= 0 ; k < 4 ; ++k) {
	    assert(((((((x * BLOCKSIZE) + i) * height)+ ((y * BLOCKSIZE) + j)) * 4) + k)<block.width*block.height*block.pixelSize) ;
	    assert(((((((y * BLOCKSIZE) + j) * width) + ((x * BLOCKSIZE) + i)) * 4) + k)<block.width*block.height*block.pixelSize) ;
	    new_image[(((((x * BLOCKSIZE) + i) * height)+ ((y * BLOCKSIZE) + j)) * 4) + k]= old_image[(((((y * BLOCKSIZE) + j) * width) + ((x * BLOCKSIZE) + i)) * 4) + k] ;
	  }
	}
      }
    }
  }
  block.width= height ;
  block.height= width ;
  block.pitch= block.pixelSize * block.width ;
  block.pixelPtr= new_image ;
#else // !BLOCKED
#ifdef IN_BLOCK_FLIP
  for (i= 0 ; i < (width*height) ; ++i) {
    new_image[i]= old_image[i*4] ;
  }
  block.width= height ;
  block.height= width ;
  block.pitch= 1 ;
  block.pixelSize= block.width ;
  block.pixelPtr= new_image ;
  block.offset[0] = 0;
  block.offset[1] = 0;
  block.offset[2] = 0;
  block.offset[3] = 0;
#else // !IN_BLOCK_FLIP
  // No method defined, so generate error!
  no_method_defined == 1 ;
#endif //IN_BLOCK_FLIP
#endif //UNBLOCKED
#endif //BLOCKED

  /* put the photo block back into the image */
#ifdef __alpha
 Tk_PhotoPutBlock(photo, &block, 0, 0, block.width, block.height);
#else
 Tk_PhotoPutBlock(photo, &block, 0, 0, block.width, block.height,TK_PHOTO_COMPOSITE_SET);
#endif
  /* Free the photo block's memory now it is finished with */
  if (block.pixelPtr) {
    Tcl_Free((char *)block.pixelPtr);
  }

  return TCL_OK;
}


