# $Id: imagedisplay.tcl,v 1.34 2010/12/17 11:24:35 ojohnson Exp $
package provide imagedisplay 3.0

set ::first_view_update 0

namespace eval img {}

itk::usual ImageDisplay { }
#itk::usual QuickDraw { }gmin\

# Create cursors for magnifier
namespace eval cursor {}
if {($::tcl_platform(os) != "Darwin") &&
    ($::tcl_platform(os) != "Windows NT")} {
    set ::cursor::box "@[file join $env(MOSFLM_GUI) bitmaps box.cursor] [file join $env(MOSFLM_GUI) bitmaps box.mask] black white"
    set ::cursor::zoom "@[file join $env(MOSFLM_GUI) bitmaps zoom.xbm] [file join $env(MOSFLM_GUI) bitmaps zoom_mask.xbm] black white"
    set ::cursor::mask "@[file join $env(MOSFLM_GUI) bitmaps mask_cursor.xbm] [file join $env(MOSFLM_GUI) bitmaps mask_mask.xbm] black white"
    set ::cursor::eraser "@[file join $env(MOSFLM_GUI) bitmaps eraser.xbm] [file join $env(MOSFLM_GUI) bitmaps eraser_mask.xbm] black pink"
} else {
    set ::cursor::box "dotbox"
    set ::cursor::zoom "dotbox"
    set ::cursor::mask "right_ptr"
    set ::cursor::eraser "cross_reverse"
}


##################################################
# Class ImageDisplay
# The class to represent the display for the image
# and all of the associated bits and bobs.
##################################################

class ImageDisplay {
    inherit itk::Toplevel

    #############################################
    # member variables
    #############################################

    # CURRENT IMAGE
	private variable prediction_nodes ""
	private variable currenthkl {}

	private variable imagegamma "1"

    # Flag to indicate if image has been opened
    private variable image_has_been_opened "0"
    # Flag to indicate if image is open
    private variable image_is_open "0"
    # Pointer to the image object currently displayed
    private variable image_object ""
    private variable next_image_object ""
    # Current template (textvariable)
    private variable current_template ""
    # Current image number (textvariable)
    private variable current_image_number ""
    private variable hide_active_mask "0"
    private variable show_info_labels "0"

    # Image lists
    private variable image_list {}

    # PANNING

    # variables to store cursor and bindings for restoration after panning
    private variable last_cursor ""
    private variable last_motion_binding ""
    # variables to store panning anchor
    private variable grab_x ""
    private variable grab_y ""
    private variable grab_left_x ""
    private variable grab_top_y ""
    private variable grab_right_x ""
    private variable grab_bottom_y ""

    # TK IMAGES

    # 3 images: master, view and zoom
    public variable master_image ""
    public variable view_image ""
    public variable zoom_image ""

    # IMAGE PROPERTIES

    # Size of master image
    private variable detector_size_x ""
    private variable detector_size_y ""
    # Size of displayed image
    private variable display_size_x ""
    private variable display_size_y ""

    # VIEW

    # Master coordinates of corners of view image
    private variable left_x 1
    private variable top_y 1
    private variable right_x 0
    private variable bottom_y 0

    # Zoom factor and min/max
    public variable zoom ""
    private variable max_zoom 64
    private variable min_zoom 0.25

    # Zoom dragging variables
    private variable zoom_click_x ""
    private variable zoom_click_y ""

    # variables to hold previous zoom and view position
    # NB. Used for plotting spots and predictions, so items can be
    # scaled and moved rather than redrawn
    private variable old_zoom ""
    private variable old_left_x 1
    private variable old_top_y 1

#     private variable zoom_queue ""
#     private variable num_zoom_clicks 0

    # INFO LABELS

    # variables to hold bindings for resotring after showing info labels
    private variable pre_info_bindings ; # array
    private variable info_cursor ""

    private variable pre_zoom_bindings ; #array
    private variable pre_magnify_bindings ; #array


    # PREDICTIONS

    private variable prediction_width 0
    private variable prediction_height 0

    # CURSORS

    # Variable storing cursor to use when not showing hourglass
    private variable current_cursor ""

    # SETTINGS

    # variables used for interactive parameter setting
    private variable origin ""
    private variable axis_order ""
    private variable two_theta_direction ""
    private variable beam_x ""
    private variable beam_y ""
    private variable backstop_x ""
    private variable backstop_y ""
    private variable backstop_radius "5.00"
    private variable search_area_min_radius ""
    private variable search_area_max_radius ""
    private variable exclusion_segment_horizontal_check "0"
    private variable exclusion_segment_horizontal "0.00"
    private variable exclusion_segment_vertical_check "0"
    private variable exclusion_segment_vertical "0.00"
    private variable orientation "North"
    private variable offset "0.00"
    private variable distance ""
    private variable wavelength ""
    private variable pixel_size ""

    private variable high_resolution_limit ""
    private variable high_resolution_radius ""
    private variable low_resolution_limit ""
    private variable low_resolution_radius ""

    private variable i_sig_i "20"

    # variables for overlay editing
    private variable current_spot_search_parameter ""
    private variable new_spot_search_parameter ""
    
    private variable old_horizontal_exclusion "0"
    private variable old_vertical_exclusion "0"

    # marker deleting
    private variable highlighted_cross ""
    private variable highlighted_cross_colour ""
    private variable highlighted_mask ""

    # INTERACTIONS

    # interactive spot finding variables
    # radius used for searching near a mouse click for
    #  spot finding/deleting
    private variable search_radius "5"
    # timer for switching spot placement mode
    private variable spot_placing_timer "2000"
    # flag to indicate spot placing mode
    private variable spot_placing_mode "auto"
    # job queue for dynamic automatic spot placing
    private variable spot_placing_queue ""
    # job queue for stippling
    private variable stipple_queue ""

#	private variable current_cursor_remember ""

    #############################################
    # methods
    #############################################

    # accessor methods
    public method getImage ;# image shown
    public method getNextImage ;# image to be shown
    public method getCanvases ;# view and zoom canvases

    # reports whether an image is shown or not
    public method isOpen

    # showing/hiding the viewer window
    public method show
    public method hide

    # enabling/disablingh the viewer window
    public method disable
    public method enable
    private method toggleAbility

    # Open an image
    public method openImage
    public method openCurrentImage

    # Close the image
    public method closeImage

    # Jump to another image
    public method gotoImage

    # Image combo methods
    public method updateImageList
    public method selectImage
    public method updateImageArrows

    # Edit spots
    public method editSpots

    # Show a new image
    public method updateImage { newImage }
    public method quickLoad
    public method updateFirstImage

    # Reset viewer size
    public method resetViewerSize

    # Refresh the view
    public method updateView

    # Interaction utility functions
    public method getMasterPixel

    # Zooming methods
    public method resetView
    public method zoomView { x1 y1 x2 y2 new_zoom}
    public method zoom
    public method zoomClick
    public method zoomDrag
    public method zoomRelease

    # panning
    public method panningClick
    public method panningRelease
    public method pan
    #public method grabClick { x y }
    #public method grabMove { x y }
    #public method grabRelease { x y }

    # Marker displaying methods
    public method plotSpots { }
    public method positionMarkers
    
    # Methods used for showing bad spots
    public method getBadSpots
    public method clearBadSpots
    public method processBadSpots
    
    private method plotBadSpots 
    
    # Methods used for showing predictions
    public method getPredictions
    public method clearPredictions
    public method processPredictions

    private method plotPredictions 
    private method plotRectangle

    # Pick methods
    public method processPick

    # Magnifier methods
    public method magnifyStart
    public method magnifyMove
    public method magnifyEnd

    # Tool toggle methods
    # overlays
    public method toggleSpots
    public method togglePredictions
    public method toggleBeam
    public method toggleMasks
    public method toggleSpotfindingParameters
    public method toggleResolutionLimits
    public method toggleActiveMask
    public method toggleInfoLabels
    # tools
    public method toggleZoomTool
    public method togglePointerTool
    public method toggleMaskingTool
    public method toggleSpotAddingTool
    public method toggleCircleFittingTool
    public method toggleEraserTool
    public method togglePanningTool

    public method toggleTools ; # obsolete

    # constant interaction bindings

    private method storePreBindings
    private method restorePreBindings

    public method startInfoLabels
    public method endInfoLabels
    public method showInfoLabels

    # tool bindings
    private method setupZoomToolBindings
    private method removeZoomToolBindings
    private method setupSpotAddingBindings
    private method removeSpotAddingBindings
    private method setupEraserBindings
    private method removeEraserBindings
    private method setupPanningBindings
    private method removePanningBindings

    private method enableZoomMagnifier
    private method disableZoomMagnifier

    private method clearCanvasBindings

    # Data access methods
    public method parseHistogram

    # NEW METHODS FOR INTERACTIVE SPOTFIND PARAM SETTING
    public method updateSetting
    public method plotBeam
    public method plotBackstop
    public method plotSpotSearchSettings
    public method plotResolutionLimits
    public method plotBackstopRadius
    public method plotBackstopCentre
    public method plotMaxSearchRadius
    public method plotMinSearchRadius
    public method plotExclusionSegment
    public method plotHighResolutionLimit
    public method plotLowResolutionLimit
    public method plotCircle
    public method plotExclusion
    public method stippleOut
    public method stippleOutSly
    public method getCurrentViewCentrePixel

    public method editSpotSearchParameterStart
    public method editSpotSearchParameterMove
    public method editSpotSearchParameterClick
    public method editSpotSearchParameterRelease
    public method editSpotSearchParameterDrag
    public method editSpotSearchParameterEnd

    # I think the following three are redundant
    public method infoPointStart
    public method infoPointMove
    public method infoPointEnd

    public method query

    # Methods used for interactive spot finding

    public method placeSpotStart
    public method placeSpotMove
    public method placeSpotClick
    public method placeSpotRelease
    public method placeSpotEnd
    public method placeSpot
    public method localSpotSearch
    public method deleteMarkerStart
    public method deleteMarkerMove
    public method deleteMarkerClick
    public method deleteMarkerRelease
    public method deleteMarkerEnd

    private method findClosestNearItem

#added by luke on 31 October 2007
	public method getCoorx
	public method getCoory
	private method startInfoLabelsScroll
	private method showInfoLabelsScroll
	private method endInfoLabelsScroll
	private method zoomluke
	private method resetViewerSizeLuke
	private method resizeWindow
	private method zoomViewLuke
	
	public method adjustContrast
#added by luke
	private method fillWindow
	private variable abs_min_zoom
	private variable luke_display_size_x
	private variable luke_display_size_y
################

	public method showFindHKLDialog
    # method for hacking!
    public method hack 

	public method findHKL
	public method plotHKL

	public method mousePick

    #############################################

    constructor { args } {

	itk_option add hull.menu

	#itk_option add hull.borderwidth
	#$itk_component(hull) configure -borderwidth 0

	wm title .image "\[No image\] - Mosflm"
	wm resizable .image 1 1
	wm iconbitmap $itk_component(hull) [wm iconbitmap .]
	wm iconmask $itk_component(hull) [wm iconmask .]
	wm group $itk_component(hull) .
	wm protocol $itk_component(hull) \
	    WM_DELETE_WINDOW [code $this hide]
	hide

    wm minsize .image 600 200

	### Menu bar and menus ####################################
	###########################################################

	itk_component add menu {
	    menu $itk_interior.menu \
		-tearoff 0 \
		-borderwidth 1
	}

	.image configure \
	    -menu $itk_component(menu)
	
	# Image menu ###############################################
	
	
	itk_component add imagemenu {
	    menu $itk_interior.menu.image \
		-tearoff false \
		-relief raised \
		-borderwidth 1 \
		-activeborderwidth 0
	}

	$itk_component(menu) add cascade \
	    -label "Image" \
	    -menu $itk_component(imagemenu)
	
	$itk_component(imagemenu) add command \
	    -label Next \
	    -underline 0 \
	    -accelerator "Alt-right" \
	    -command [code $this gotoImage "next"]
    
	$itk_component(imagemenu) add command \
	    -label Preceding \
	    -underline 0 \
	    -accelerator "Alt-left" \
	    -command [code $this gotoImage "prev"]
    
	bind $itk_component(hull) <Alt-Right> [code $this gotoImage "next"]
	bind $itk_component(hull) <Alt-Left> [code $this gotoImage "prev"]


    # Add menu icons on linux

	if {($::tcl_platform(os) != "Darwin") &&
	    ($::tcl_platform(os) != "Windows NT")} {
	    
	    $itk_component(imagemenu) entryconfigure 0 \
		-compound left \
		-image ::img::next_arrow

	    $itk_component(imagemenu) entryconfigure 1 \
		-compound left \
		-image ::img::prev_arrow

	}

	
	
	# View menu ###############################################
	
# 	itk_component add viewmenubutton {
# 	    menubutton $itk_interior.menubar.view \
# 		-text View \
# 		-underline 0 \
# 		-borderwidth 1 \
# 		-state disabled
# 	}
# 	pack $itk_component(viewmenubutton) -side left
	
	itk_component add viewmenu {
	    menu $itk_interior.menu.view \
		-tearoff false \
		-relief raised \
		-borderwidth 1 \
		-activeborderwidth 0
	}
# 	$itk_component(viewmenubutton) configure \
# 	    -menu $itk_component(viewmenu)
	
	$itk_component(menu) add cascade \
	    -label "View" \
	    -menu $itk_component(viewmenu)

	itk_component add toolsmenu {
		menu $itk_interior.menu.tools \
		-tearoff false \
		-relief raised \
		-borderwidth 1 \
		-activeborderwidth 0
	}	

	$itk_component(menu) add cascade \
		-label "Tools" \
		-menu $itk_component(toolsmenu)

	$itk_component(toolsmenu) add command \
	    -label "Find HKL" \
	    -underline 0 \
	    -command [code $this showFindHKLDialog]

	# Tools menu ##############################################

# 	itk_component add toolsmenubutton {
# 	    menubutton $itk_interior.menubar.tools \
# 		-text Tools \
# 		-underline 0 \
# 		-borderwidth 1 \
# 		-state disabled
# 	}
# 	pack $itk_component(toolsmenubutton) -side left
	
# 	itk_component add toolsmenu {
# 	    menu $itk_interior.menubar.tools.menu \
# 		-tearoff false \
# 		-relief raised \
# 		-borderwidth 1 \
# 		-activeborderwidth 0
# 	}
# 	$itk_component(toolsmenubutton) configure \
# 	    -menu $itk_component(toolsmenu)
	
# 	$itk_component(toolsmenu) add command \
# 	    -label "Find spots" \
# 	    -command [code $this findSpots] \
# 	    -state disabled
	
	# Tool bar components #####################################
	###########################################################

	# Toolbar rows
	
	itk_component add toolbar_row_1 {
	    frame $itk_interior.tbr1 \
		    -relief flat \
		    -borderwidth 0
	} {
	    ignore -background
	}
	pack $itk_component(toolbar_row_1) -side top -fill x -expand 0
	
	itk_component add toolbar_row_2 {
	    frame $itk_interior.tbr2 \
		    -relief flat \
		    -borderwidth 0
	}
	pack $itk_component(toolbar_row_2) -side top -fill x -expand 0
	
	itk_component add toolbar_row_3 {
	    frame $itk_interior.tbr3 \
		    -relief flat \
		    -borderwidth 0
	}
	#pack $itk_component(toolbar_row_3) -side top -fill x -expand 1
	
	# Image toolbar

	itk_component add image_toolbar {
	    frame $itk_component(toolbar_row_1).itb \
		    -relief raised \
		    -borderwidth 1
	} {
	    keep -background
	}
	pack $itk_component(image_toolbar) -side left

	itk_component add preceding_image {
	    Toolbutton $itk_component(image_toolbar).prev \
		-image ::img::prev_arrow \
		-disabledimage ::img::prev_arrow_disabled \
		-type "amodal" \
		-state "normal" \
		-balloonhelp " Preceding image " \
		-command [code $this gotoImage "prev"]
	}
	pack $itk_component(preceding_image) -side left -padx 1
	
	itk_component add next_image {
	    Toolbutton $itk_component(image_toolbar).next \
		-image ::img::next_arrow \
		-disabledimage ::img::next_arrow_disabled \
		-type "amodal" \
		-state "normal" \
		-balloonhelp " Next image " \
		-command [code $this gotoImage "next"]
	}
	pack $itk_component(next_image) -side left -padx 1
		
	itk_component add image_combo {
	    combobox::combobox $itk_component(image_toolbar).ic \
		-width 40 \
		-editable 0 \
		-highlightcolor black \
		-command [code $this selectImage]
	} {
	    keep -background -cursor -foreground -font
	    keep -selectbackground -selectborderwidth -selectforeground
	    keep -highlightcolor -highlightthickness
	    rename -highlightbackground -background background Background
	    rename -background -textbackground textBackground Background
	}
	pack $itk_component(image_combo) -side left -padx 1

	# markers toolbar
		
	itk_component add markers_toolbar {
	    frame $itk_component(toolbar_row_2).mtb \
		    -relief raised \
		    -borderwidth 1
	}
	pack $itk_component(markers_toolbar) -side left 
				
	itk_component add beam {
	    Toolbutton $itk_component(markers_toolbar).beam \
		-image ::img::beam16x16 \
		-activeimage ::img::beam_on16x16 \
		-disabledimage ::img::beam16x16 \
		-state "normal" \
		-type "modal" \
		-balloonhelp " Show beam centre " \
		-command [code $this toggleBeam]
	}
	pack $itk_component(beam) -side left -padx 1

	itk_component add backstop {
	    Toolbutton $itk_component(markers_toolbar).backstop \
		-image ::img::backstop16x16 \
		-state "normal" \
		-type "modal" \
		-balloonhelp " Show backstop centre " \
		-command [code $this toggleBackstop]
	}
	#pack $itk_component(backstop) -side left -padx 1

	itk_component add spots {
	    Toolbutton $itk_component(markers_toolbar).spots \
		-image ::img::spot16x16 \
		-activeimage ::img::spot_on16x16 \
		-disabledimage ::img::spot16x16 \
		-state "normal" \
		-type "modal" \
 		-balloonhelp " Show spots " \
		-command [code $this toggleSpots]
	}
	pack $itk_component(spots) -side left -padx 1
	
	itk_component add predictions {
	    Toolbutton $itk_component(markers_toolbar).prediction \
		-image ::img::boxes16x16 \
		-activeimage ::img::boxes_on16x16 \
		-disabledimage ::img::boxes_disabled16x16 \
		-state "disabled" \
		-type "modal" \
		-balloonhelp " Show predictions " \
		-command [code $this togglePredictions]
	}
	pack $itk_component(predictions) -side left -padx 1
	
	itk_component add masks {
	    Toolbutton $itk_component(markers_toolbar).masks \
		-image ::img::masks16x16 \
		-activeimage ::img::masks_on16x16 \
		-disabledimage ::img::masks16x16 \
		-type "modal" \
		-state "normal" \
		-balloonhelp " Show masked areas (including backstop)" \
		-command [code $this toggleMasks]
	}
	pack $itk_component(masks) -side left -padx 1
	
	itk_component add spotfinding_settings {
	    Toolbutton $itk_component(markers_toolbar).spotfinding \
		-image ::img::spot_search_area16x16 \
		-activeimage ::img::spot_search_area_on16x16 \
		-state "normal" \
		-type "modal" \
		-balloonhelp " Show spotfinding search area " \
		-command [code $this toggleSpotfindingParameters]
	}
	pack $itk_component(spotfinding_settings) -side left -padx 1

	itk_component add resolution_limits {
	    Toolbutton $itk_component(markers_toolbar).resolution \
		-image ::img::resolution_limits16x16 \
		-activeimage ::img::resolution_limits_on16x16 \
		-state "normal" \
		-type "modal" \
		-balloonhelp " Show resolution limits " \
		-command [code $this toggleResolutionLimits]
	}
	pack $itk_component(resolution_limits) -side left -padx 1

	itk_component add active_mask {
	    Toolbutton $itk_component(markers_toolbar).actmsk \
		-image ::img::rigaku16x16 \
		-activeimage ::img::rigaku16x16 \
		-disabledimage ::img::rigaku_grey16x16 \
		-state "normal" \
		-type "modal" \
 		-balloonhelp " Hide active mask on Rigaku detector " \
		-command [code $this toggleActiveMask]
	}
	pack $itk_component(active_mask) -side left -padx 1

	itk_component add info_labels {
	    Toolbutton $itk_component(markers_toolbar).il \
		-image ::img::info_labels \
		-disabledimage ::img::info_labels \
		-type "modal" \
		-state "normal" \
		-balloonhelp " Show info labels " \
		-command [code $this toggleInfoLabels]
	}
# 	pack $itk_component(info_labels) -side left -padx 1


	# Tools toolbar

	itk_component add tools_toolbar {
	    frame $itk_component(toolbar_row_2).ttb \
		    -relief raised \
		    -borderwidth 1
	}
	pack $itk_component(tools_toolbar) -side left 
	
	itk_component add zoom_tool {
	    Toolbutton $itk_component(tools_toolbar).zoomtool \
		-image ::img::magnifier \
		-activeimage ::img::magnifier_on \
		-disabledimage ::img::magnifier \
		-type "radio" \
		-group "image_tools" \
		-state "normal" \
		-balloonhelp " Zoom tool " \
		-command [code $this toggleZoomTool]
	}
	pack $itk_component(zoom_tool) -side left -padx 1
	
	itk_component add panning_tool {
	    Toolbutton $itk_component(tools_toolbar).pt \
		-image ::img::fleur16x16 \
		-activeimage ::img::fleur_on16x16 \
		-disabledimage ::img::fleur16x16 \
		-type "radio" \
		-group "image_tools" \
		-state "normal" \
		-balloonhelp " Panning tool " \
		-command [code $this togglePanningTool]
	}
 	pack $itk_component(panning_tool) -side left -padx 1

	itk_component add pointer {
	    Toolbutton $itk_component(tools_toolbar).pointer \
		-image ::img::pointer16x16 \
		-activeimage ::img::pointer_on16x16 \
		-disabledimage ::img::pointer16x16 \
		-type "radio" \
		-group "image_tools" \
		-state "normal" \
		-balloonhelp " Selection tool for editing search area, processing limits and masks " \
		-command [code $this togglePointerTool]
	}
	pack $itk_component(pointer) -side left -padx 1
		
	itk_component add spotfinding {
	    Toolbutton $itk_component(tools_toolbar).spotfinding \
		-image ::img::pencil16x16 \
		-activeimage ::img::pencil_on16x16 \
		-disabledimage ::img::pencil16x16 \
		-type "radio" \
		-group "image_tools" \
		-state "normal" \
		-balloonhelp " Manually add spots " \
		-command [code $this toggleSpotAddingTool]
	}
	pack $itk_component(spotfinding) -side left -padx 1
	
	itk_component add masking_tool {
	    Toolbutton $itk_component(tools_toolbar).mt \
		-image ::img::masking16x16 \
		-activeimage ::img::masking_on16x16 \
		-disabledimage ::img::masking16x16 \
		-type "radio" \
		-group "image_tools" \
		-state "normal" \
		-balloonhelp " Masking tool " \
		-command [code $this toggleMaskingTool]
	}
 	pack $itk_component(masking_tool) -side left -padx 1

	itk_component add circle_fitting {
	    Toolbutton $itk_component(tools_toolbar).cf \
		-image ::img::circle_fitting16x16 \
		-activeimage ::img::circle_fitting_on16x16 \
		-disabledimage ::img::circle_fitting16x16 \
		-type "radio" \
		-group "image_tools" \
		-state "normal" \
		-balloonhelp " Circle fitting " \
		-command [code $this toggleCircleFittingTool]
	}
 	pack $itk_component(circle_fitting) -side left -padx 1

	itk_component add spotdeleting {
	    Toolbutton $itk_component(tools_toolbar).spotdeleting \
		-image ::img::eraser16x16 \
		-activeimage ::img::eraser_on16x16 \
		-disabledimage ::img::eraser16x16 \
		-type "radio" \
		-group "image_tools" \
		-state "normal" \
		-balloonhelp " Spots and mask eraser " \
		-command [code $this toggleEraserTool]
	}
	pack $itk_component(spotdeleting) -side left -padx 1
	
	# Zoom toolbar

	itk_component add zoom_toolbar {
	    frame $itk_interior.tbr1.ztb \
		-relief raised \
		-borderwidth 1
	}
	pack $itk_component(zoom_toolbar) -side left 
		
	
	itk_component add zoom_in_tool {
	    Toolbutton $itk_component(zoom_toolbar).zit \
		-image ::img::zoom_in16x16 \
		-disabledimage ::img::zoom_in_disabled16x16 \
		-type "amodal" \
		-state "normal" \
		-balloonhelp " Zoom in " \
		-command [code $this zoom "in"]
	}
	pack $itk_component(zoom_in_tool) -side left -padx 1
	
	itk_component add zoom_out_tool {
	    Toolbutton $itk_component(zoom_toolbar).zot \
		-image ::img::zoom_out16x16 \
		-disabledimage ::img::zoom_out_disabled16x16 \
		-type "amodal" \
		-state "normal" \
		-balloonhelp " Zoom out " \
		-command [code $this zoom "out"]
	}
	pack $itk_component(zoom_out_tool) -side left -padx 1
	
	itk_component add fit_image {
	    Toolbutton $itk_component(zoom_toolbar).fi \
		-image ::img::fit_image16x16 \
		-disabledimage ::img::fit_image16x16 \
		-type "amodal" \
		-state "normal" \
		-balloonhelp " Fit image " \
		-command [code $this resetViewerSizeLuke]
#The abobe command used to be resetView but I changed it to resetViewerSizeLuke
	}
	pack $itk_component(fit_image) -side left -padx 1
	
	itk_component add contrast_button {
	    Toolbutton $itk_component(zoom_toolbar).contrast \
		-image ::img::contrast16x16 \
		-disabledimage ::img::contrast16x16 \
		-type "modal" \
		-state "normal" \
		-balloonhelp " Contrast "
	}
	pack $itk_component(contrast_button) -side left -padx 1

	itk_component add contrast_entry {
		entry $itk_component(zoom_toolbar).contrastentry
	}
#	pack $itk_component(contrast_entry) -side left -padx 1
	bind $itk_component(contrast_entry) <Return> [code $this adjustContrast]
	
	itk_component add contrast_spin {
		spinbox $itk_component(zoom_toolbar).contrastspin \
		-from 0.00 -to 5.00 \
		-increment 0.5 -format %5.2f -width 10
	} {
		keep -background
	}
#	pack $itk_component(contrast_spin) -side left -padx 1
#	bind $itk_component(contrast_entry) <Return> [code $this adjustContrast]


	itk_component add contrast_palette {
	    Contrast .contrast_palette
	}
	$itk_component(contrast_button) configure \
	    -command [list $itk_component(contrast_palette) launch $itk_component(contrast_button)]
	
	# filler toolbars

	itk_component add toolbarX1 {
	    frame $itk_component(toolbar_row_1).tbx \
		    -relief raised \
		    -borderwidth 1
	}
	pack $itk_component(toolbarX1) -side left -fill both -expand 1
		
	itk_component add toolbarX2 {
	    frame $itk_component(toolbar_row_2).tbx \
		    -relief raised \
		    -borderwidth 1
	}
	pack $itk_component(toolbarX2) -side left -fill both -expand 1
		
	itk_component add toolbarX3 {
	    frame $itk_component(toolbar_row_3).tbx \
		    -relief raised \
		    -borderwidth 1
	}
	pack $itk_component(toolbarX3) -side right -fill both -expand 1


	itk_component add main {
	    frame $itk_interior.main \
		    -relief raised \
		    -borderwidth 1 \
		    -width 502 \
		    -height 502
	} {
	    keep -background
	}
	pack $itk_component(main) -fill both -expand true

	# Canvas component ########################################
	###########################################################

	itk_component add canvas {
	    canvas $itk_interior.main.canvas \
		-width 500 \
		-height 500 \
		-bg #a9a9a9 \
		-relief sunken \
		-bd 0 \
		-highlightthickness 0
	}
	pack $itk_component(canvas) -expand true


	bind $itk_component(main) <Configure> [code $this resizeWindow]

	# intialize cursor for initial tool (zoom)
	if {[tk windowingsystem] == "x11"} {
	    $itk_component(canvas) configure -cursor $::cursor::zoom
	}	    

	# Magnifier component #####################################
	###########################################################
	
	itk_component add magnifier {
	    toplevel .magnifier -bg #a9a9a9 -borderwidth 0
	} {
	}
	wm overrideredirect $itk_component(magnifier) 1
	if {[tk windowingsystem] == "aqua"} {
	    ::tk::unsupported::MacWindowStyle style $itk_component(magnifier) help none 
	}
	wm withdraw $itk_component(magnifier)
	
	itk_component add border {
	    frame .magnifier.border \
		    -bg #000000
	} {
	}
	pack $itk_component(border)

	itk_component add mag_canvas {
	    canvas .magnifier.border.c \
		    -width 80\
		    -height 80\
		    -bg #a9a9a9 \
		    -relief sunken \
		    -bd 0 \
		    -highlightthickness 0 \
		    -highlightbackground #dcdcdc \
		    -cursor "left_ptr" \
		    -xscrollincrement 1 \
		    -yscrollincrement 1
	}
	pack $itk_component(mag_canvas) -padx 1 -pady 1 -fill both -expand 1
	
	# Info labels component ########################################
	################################################################

	itk_component add data_labels {
	    toplevel .dataLabel \
		-border 0
	}
	wm overrideredirect $itk_component(data_labels) 1
	if {[tk windowingsystem] == "aqua"} {
	    ::tk::unsupported::MacWindowStyle style $itk_component(data_labels) help none
	}
	wm withdraw $itk_component(data_labels)

	itk_component add data_label_1 {
	    label .dataLabel.l1 \
		-text "?"\
		-relief solid \
		-highlightthickness 0 \
		-bd 2 \
		-font font_b
	} {
	    usual
	    ignore -font -highlightthickness
	}
	pack $itk_component(data_label_1) -fill x

	itk_component add data_label_2 {
	    label .dataLabel.l2 \
		-text "?"\
		-relief solid \
		-highlightthickness 0 \
		-bd 2 \
		-font font_b
	} {
	    usual
	    ignore -font -highlightthickness
	}
	itk_component add data_label_3 {
	    label .dataLabel.l3 \
		-text "?"\
		-relief solid \
		-highlightthickness 0 \
		-bd 2 \
		-font font_b
	} {
	    usual
	    ignore -font -highlightthickness
	}
	itk_component add data_label_4 {
	    label .dataLabel.l4 \
		-text "?"\
		-relief solid \
		-highlightthickness 0 \
		-bd 2 \
		-font font_b
	} {
	    usual
	    ignore -font -highlightthickness
	}
	itk_component add data_label_5 {
	    label .dataLabel.l5 \
		-text "?"\
		-relief solid \
		-highlightthickness 0 \
		-bd 2 \
		-font font_b
	} {
	    usual
	    ignore -font -highlightthickness
	}
	
	# Overlays #####################################################
	################################################################

	Overlay::initialize $itk_component(canvas)

	# Header component #############################################
	################################################################
	
	# Create images (to be filled with data on demand) #############
	set master_image [image create photo]
	set view_image [image create photo]
	set zoom_image [image create photo]

	# Display the images in their canvases #########################
	$itk_component(canvas) create image 0 0 \
	    -image $view_image -tags view_image -anchor nw 
	$itk_component(mag_canvas) create image 0 0 \
	    -image $zoom_image -anchor nw -tags zoom_image

	# Plot the origin on the mag canvas ############################
	$itk_component(mag_canvas) create text 0 0 -text "X" -tags [list origin marker]


	# Evalutate options ############################################
	eval itk_initialize $args

	set ::image_open 0

    }
}

body ImageDisplay::hack {  } {
}

body ImageDisplay::mousePick {a_x a_y} {
	if {$zoom != $max_zoom} { 
	set l_pixel_coords [getMasterPixel $a_x $a_y]
	set pixel_coord_x [lindex $l_pixel_coords 0]
	set pixel_coord_y [lindex $l_pixel_coords 1]
	while {[expr $pixel_coord_x - 5] < 1} {
		incr pixel_coord_x
	}

	while {[expr $pixel_coord_y - 5] < 1} {
		incr pixel_coord_y
	}


	while {[expr $pixel_coord_y - 5] < 1} {
		incr pixel_coord_y
	}

	while {[expr $pixel_coord_y + 5] > [expr [$::session getImageHeight] - 1]} {
		incr pixel_coord_y -1
	}

	while {[expr $pixel_coord_x + 5] > [expr [$::session getImageWidth] - 1]} {
		incr pixel_coord_x -1
	}

	set mp_image_number [$image_object getNumber]

	$::mosflm sendCommand "pick $mp_image_number $pixel_coord_x $pixel_coord_y 11 11 \n"
	$::mosflm sendCommand "go \n"
#	puts "pick $mp_image_number $pixel_coord_x $pixel_coord_y 11 11"
	} 
}

body ImageDisplay::adjustContrast {} {
	set gammavalue [$itk_component(contrast_entry) get]
	set dest [image create photo -gamma 1]
	$dest copy $view_image
	image delete $view_image
	set view_image [image create photo -gamma $gammavalue]
	$view_image copy $dest
	$itk_component(canvas) itemconfigure view_image -image $view_image
	image delete $dest
}

##############################################################################
# Accessor methods
##############################################################################

body ImageDisplay::getImage { } {
    return $image_object
}

body ImageDisplay::getNextImage { } {
    return $next_image_object
}

body ImageDisplay::getCanvases { } {
    return [list $itk_component(canvas) $itk_component(mag_canvas)]
}

##############################################################################
# Method to report whther an image is open 
##############################################################################

body ImageDisplay::isOpen { } {
    return $image_is_open
}

##############################################################################
# Methods to show and hide image viewer 
##############################################################################

body ImageDisplay::show { } {
    wm deiconify $itk_component(hull)
}

body ImageDisplay::hide { } {
    wm withdraw $itk_component(hull)
}

##############################################################################
# enabling/disablingh the viewer window
##############################################################################

body ImageDisplay::disable { } {
    toggleAbility "disabled"
    $itk_component(preceding_image) configure -state disabled
    $itk_component(next_image) configure -state disabled
}

body ImageDisplay::enable { } {
    toggleAbility "normal"
	updateImageArrows
}

body ImageDisplay::toggleAbility { a_state } {
    $itk_component(menu) entryconfigure 0 -state $a_state
    $itk_component(image_combo) configure -state $a_state
    $itk_component(contrast_button) configure -state $a_state

}

##############################################################################
# Method to open a new image in the viewer
##############################################################################

body ImageDisplay::openImage { an_image } {


    # disable viewer 
    .image disable

    # Update image controls and title
    set next_image_object  $an_image

    # Getcontrast limits
    foreach { l_min l_max } [$itk_component(contrast_palette) getContrast] break
    
    # Get the image from mosflm
    $::mosflm getImage $an_image $l_min $l_max
    
    # If there is an a-matrix, get predictions
    if {[[$an_image getSector] reportMatrix] != "Unknown"} {
#	puts "OPENIMAGE"
	getPredictions
    }

    # N.B. receipt of image data and updateImage

}    

body ImageDisplay::openCurrentImage { an_image } {

    # Update image controls and title
    set next_image_object $an_image

    # Getcontrast limits
    foreach { l_min l_max } [$itk_component(contrast_palette) getContrast] break
    
    # Get the current image from mosflm
    $::mosflm getCurrentImage $l_min $l_max
    # Get the current predictions
    #puts "OPENCURRENTIMAGE and getCurrentPredictions"
    $::mosflm getCurrentPredictions

    # N.B. receipt of image data and updateImage

}    

body ImageDisplay::closeImage { } {
    set image_object ""
    wm title .image "No image displayed - Mosflm"
    hide
    #clearTemplates

    # Clear master image
    image delete $master_image
    set master_image [image create photo]

    # Reset view image
    image delete $view_image
    set view_image [image create photo]
    $itk_component(canvas) itemconfigure view_image -image $view_image

    # Reset the contrast palette
    $itk_component(contrast_palette) reset

    # Set flag to indicate no image has been opened in current session
    set image_has_been_opened 0

    # Turn off image bindings
    foreach i_event [bind $itk_component(canvas)] {
	bind $itk_component(canvas) $i_event {}
    }

    # Disable image combo
    $itk_component(image_combo) configure -state disabled

    # Turn off overlays
    foreach i_toolbutton {spots predictions spotfinding_settings resolution_limits info_labels spotfinding spotdeleting zoom_tool pointer } {
	$itk_component($i_toolbutton) cancel
    }

    # Disable tools
    foreach i_toolbutton {preceding_image next_image image_combo spots predictions spotfinding_settings resolution_limits info_labels spotfinding spotdeleting zoom_tool pointer zoom_in_tool zoom_out_tool fit_image contrast_button } {
	$itk_component($i_toolbutton) configure -state disabled
    }

    # Reset initialize flag
    set image_has_been_opened 0
}

body ImageDisplay::gotoImage { which } {
    #puts $image_list
    set curr_posn [lsearch $image_list $image_object]
    #puts "At posn. $curr_posn $image_object"
# not always moving thro sequential images so do not incr/decr image no., rather the position in list
    if {$which == "prev"} {
	set new_posn [expr $curr_posn - 1]
	set l_new_image_num [[lindex $image_list $new_posn] getNumber]
    } elseif {$which == "next"} {
	set new_posn [expr $curr_posn + 1]
	set l_new_image_num [[lindex $image_list $new_posn] getNumber]
    } else {
	error "Bad argument passed to gotoImage: $which"
    }
    #puts "New posn. $new_posn New image no. $l_new_image_num"
    set l_template [[lindex $image_list $new_posn] getTemplate]
    set l_new_image [$::session getImageByTemplateAndNumber $l_template $l_new_image_num]
    if {$l_new_image != ""} {
	openImage $l_new_image
    }
}

body ImageDisplay::updateImageList { } {
    # Get new iamge list from session
    set image_list [$::session getImages]
    # Update combo box's listvar
    set l_image_names {}
    foreach i_image $image_list {
	lappend l_image_names [$i_image getShortName]
    }
    $itk_component(image_combo) list delete 0 end
    eval $itk_component(image_combo) list insert 0 $l_image_names
    # Update image arrows
    updateImageArrows
}

body ImageDisplay::selectImage { a_combo a_filename } {
    set l_selected_image_object [$::session getImageByName $a_filename]
    if {$l_selected_image_object != $image_object} {
	openImage $l_selected_image_object
    }
}

body ImageDisplay::updateImageArrows { } {
    if {$image_object != ""} {
	set l_pos [lsearch $image_list $image_object]
	if {$l_pos > 0} {
	    # enable previous arrow
	    $itk_component(preceding_image) configure -state "normal"
	    $itk_component(imagemenu) entryconfigure 1 -state "normal"
	} else {
	    # disable previous arrow
	    $itk_component(preceding_image) configure -state "disabled"
	    $itk_component(imagemenu) entryconfigure 1 -state "disabled"
	}
	if {$l_pos < [expr [llength $image_list] -1]} {
	    # enable next arrow
	    $itk_component(next_image) configure -state "normal"
	    $itk_component(imagemenu) entryconfigure 0 -state "normal"
	} else {
	    # disable next arrow
	    $itk_component(next_image) configure -state "disabled"
	    $itk_component(imagemenu) entryconfigure 0 -state "disabled"
	}
    }
}

##############################################################################
# Method to edit spots on an image in the viewer
##############################################################################

body ImageDisplay::editSpots { { an_image "" } } {

    # if necessary open the image
    if {($an_image != "") && ($an_image != $image_object)} {
	openImage $an_image
    }

    # Select the spot adding tool
    $itk_component(spotfinding) invoke
}    

##############################################################################
# Method to dispay received jpeg
##############################################################################

body ImageDisplay::updateFirstImage { } {
    # Calculate detector size
    set detector_size_x [image width $master_image]
    set detector_size_y [image height $master_image]

    set l_screen_height [winfo screenheight $itk_component(canvas)]

    set l_screen_gubbins 200
    set l_max_space_per_pixel [expr double($l_screen_height - $l_screen_gubbins) / $detector_size_y]
    set min_zoom [expr pow(2,floor(log($l_max_space_per_pixel)/log(2)))]
	
    set abs_min_zoom $min_zoom

    # Set up view menu
    $itk_component(viewmenu) delete 0 end
    $itk_component(viewmenu) add radiobutton \
	-variable [scope min_zoom] \
	-value [expr $min_zoom / 2] \
	-label "Too small" \
	-command [code $this resetViewerSize [expr $min_zoom / 2]]
    $itk_component(viewmenu) add radiobutton \
	-variable [scope min_zoom] \
	-value $min_zoom \
	-label "Just right" \
	-command [code $this resetViewerSize $min_zoom]
    $itk_component(viewmenu) add radiobutton \
	-variable [scope min_zoom] \
	-value [expr $min_zoom * 2] \
	-label "Too big" \
	-command [code $this resetViewerSize [expr $min_zoom * 2]]

    # Rest the image viewer size
    resetViewerSize

    # Set up canvas bindings for zooming and magnifying
    $itk_component(zoom_tool) invoke
    set current_cursor $::cursor::zoom
#    $itk_component(pointer) invoke
#    set current_cursor left_ptr
    #bind $itk_component(canvas) <ButtonPress-2> [code $this grabClick %x %y]
	# added by luke 13 November 2007
	# bug 87: middle mouse button reset view on mac whereas right mouse button does
	# the same job on all other platforms. It seems that button 2 is the right button 
	# on the mac whereas button 3 is the right button on linux and windows.
	# The same statement occurs one more time in this file
    if {[tk windowingsystem] == "aqua"} {
		bind $itk_component(canvas) <ButtonPress-2> [code $this resetViewerSizeLuke]
	} else {
		bind $itk_component(canvas) <ButtonPress-3> [code $this resetViewerSizeLuke]
#the above line used to be resetView but I changed it to resetViewerSizeLuke
	}

    # Setup info labels binding
    if {[tk windowingsystem] != "aqua"} {
	bind $itk_component(canvas) <Alt-Motion> [code $this startInfoLabelsScroll "Motion" %x %y]
#	bind $itk_component(canvas) <M2-Motion> [code $this startInfoLabels "Motion" %x %y]
	bind $itk_component(canvas) <Control-Motion> [code $this startInfoLabels "Motion" %x %y]
    } else {
	bind $itk_component(canvas) <Alt-Motion> [code $this startInfoLabelsScroll "Motion" %x %y]
	bind $itk_component(canvas) <M2-Motion> [code $this startInfoLabels "Motion" %x %y]
	#bind $itk_component(canvas) <M2-1> [code $this startInfoLabels "ButtonPress-1" %x %y]
    }
    # Activate spots button here ready for spot finding
	$itk_component(spots) invoke
    #$itk_component(toolsmenu) entryconfigure 0 -state normal
    #$itk_component(canvas) configure -cursor $::cursor::target

    # Put image on canvas
    $itk_component(canvas) create image 0 0 \
	-image $view_image -tags view_image -anchor nw 
#added by luke on 12 November 2007
#The line below forces the geometry of the window to be set to 
#the dimensions that are calculated when the first image is fired up
#Without this statement the window resizes to the max when the user
#chooses to zoom in and I didn't like this behaviour
	wm geometry .image [winfo reqwidth .image]x[winfo reqheight .image]
#####################################################################
}

body ImageDisplay::quickLoad { an_image { a_min 0 } { a_max 0 } } {

    set image_object $next_image_object

    # update image combo 
    $itk_component(image_combo) configure -state normal -editable 1
    $itk_component(image_combo) delete 0 end
    $itk_component(image_combo) insert 0 [$image_object getShortName]
    $itk_component(image_combo) configure -editable 0

    # Update image changing tools (number limits + arrows)
    updateImageArrows

    # Enable tools
    foreach i_toolbutton { spots spotfinding_settings resolution_limits info_labels spotfinding spotdeleting zoom_tool pointer fit_image contrast_button } {
	$itk_component($i_toolbutton) configure -state normal
    }

    # Update the window title to reflect its content
    wm title .image "Image [$image_object getNumber] - Mosflm"

    set ::stopwatch2 [clock clicks -milli]
    # Called when jpeg recieved from server

    # Remove markers from the canvases too
    $itk_component(canvas) delete marker

    # Build an actual-size image of the whole detector from the jpeg
    set ::timer002 [clock clicks -milli]
    image delete $master_image
    set master_image [image create photo]
    set l_histo [imageLoad "ADSC" [$image_object getFullPathName] $master_image $a_min $a_max]

    set ::timer003 [clock clicks -milli]

    if {!$image_has_been_opened} {
	updateFirstImage
	set image_has_been_opened 1
    }

    # Update view (and magnifying) images 
    set ::first_view_update 1
    updateView

    # Plot spotlist if available
    plotSpots

    if {[info exists ::stopwatch3]} {
	set ::stopwatch4 [clock clicks -milli]
    }

    # Deiconfiy the window if it is withdrawn
    if {[wm state $itk_component(hull)] == "withdrawn"} {
	wm deiconify $itk_component(hull)
    }

    # Update the contrast histogram
    $itk_component(contrast_palette) updateHistogram $l_histo

    # enable image viewer
    enable
}

body ImageDisplay::updateImage { new_image_data } {

    set image_object $next_image_object

    # update image combo
    $itk_component(image_combo) configure -state normal -editable 1
    $itk_component(image_combo) delete 0 end 
    $itk_component(image_combo) insert 0 [$image_object getShortName]
    $itk_component(image_combo) configure -editable 0

    # Update image changing tools (number limits + arrows)
    updateImageArrows

    # Enable tools
    foreach i_toolbutton { spots spotfinding_settings resolution_limits info_labels spotfinding spotdeleting zoom_tool pointer fit_image contrast_button } {
	$itk_component($i_toolbutton) configure -state normal
    }

    # Update the window title to reflect its content
    foreach { l_phi_start l_phi_end } [$image_object getPhi] break
    wm title .image "Image [$image_object getNumber] (\u03c6:$l_phi_start-$l_phi_end) - Mosflm"

    set ::stopwatch2 [clock clicks -milli]
    # Called when jpeg received from server

    # Remove markers from the canvases too
    $itk_component(canvas) delete marker

    # Build an actual-size image of the whole detector from the jpeg
    $master_image put $new_image_data
    set ::timer003 [clock clicks -milli]

    if {!$image_has_been_opened} {
	updateFirstImage
	set image_has_been_opened 1
    }

    # Update view (and magnifying) images 
    set ::first_view_update 1
    updateView

    # Plot spotlist if available
    plotSpots

    if {[info exists ::stopwatch3]} {
	set ::stopwatch4 [clock clicks -milli]
    }

    # Deiconfiy the window if it is withdrawn
    if {[wm state $itk_component(hull)] == "withdrawn"} {
	wm deiconify $itk_component(hull)
    }

    # enable image viewer
    enable

    .c enableIndexing
}


body ImageDisplay::resetViewerSizeLuke { { a_zoom ""} } {

    set current_cursor_remember [$itk_component(canvas) cget -cursor]

    $itk_component(canvas) delete marker

    # Initialise position records, and previous positoin records for use in
    #  marker positioning
    set left_x 1
    set top_y 1
    set old_left_x 1
    set old_top_y 1
    set right_x $detector_size_x
    set bottom_y $detector_size_y
    set zoom $min_zoom
    set old_zoom $zoom

    # Calculate desired canvas size (To fit image at 1:4 resolution)
    set display_size_x [expr $detector_size_x * $min_zoom]
    set display_size_y [expr $detector_size_y * $min_zoom]

    # Set the canvas size
    $itk_component(canvas) configure \
	-width $display_size_x \
	-height $display_size_y
	
	update

	set cnvsfrm_width [winfo width $itk_component(main)]
	set cnvsfrm_height [winfo height $itk_component(main)]	

	if {$display_size_x > $cnvsfrm_width} {
		set display_size_x $cnvsfrm_width
	}
	
	if {$display_size_y > $cnvsfrm_height} {
		set display_size_y $cnvsfrm_height
	}

    $itk_component(canvas) configure \
	-width $display_size_x \
	-height $display_size_y

	
    # Redisplay spots
    plotSpots

    # Redisplay predictions
    $::session updatePredictions

    # Refresh view
    updateView

    $itk_component(canvas) configure \
	-scrollregion "0 0 $display_size_x $display_size_y" 

    zoomluke "in"
    zoomluke "out"
    plotSpots

    $itk_component(zoom_out_tool) configure -state "disabled"

    $itk_component(canvas) configure -cursor $current_cursor_remember
    set current_cursor $current_cursor_remember

}


body ImageDisplay::resizeWindow {} {

    $itk_component(canvas) delete marker

    set zoom $zoom
    set old_zoom $zoom

    # Calculate desired canvas size (To fit image at 1:4 resolution)

    set display_size_x [expr $detector_size_x * $zoom]
    set display_size_y [expr $detector_size_y * $zoom]


    # Set the canvas size
    $itk_component(canvas) configure \
	-width $display_size_x \
	-height $display_size_y
	
	update
	
#l	puts $zoom

#	puts [winfo reqwidth $itk_component(canvas)]
#	puts [winfo reqheight $itk_component(canvas)]
	set cnvsfrm_width [winfo width $itk_component(main)]
	set cnvsfrm_height [winfo height $itk_component(main)]	
#added by luke on 7
#    set right_x $cnvsfrm_height
#    set bottom_y $cnvsfrm_width
#################################

	if {$display_size_x > $cnvsfrm_width} {
		set display_size_x $cnvsfrm_width
	}
	
	if {$display_size_y > $cnvsfrm_height} {
		set display_size_y $cnvsfrm_height
	}

    $itk_component(canvas) configure \
	-width $display_size_x \
	-height $display_size_y

    $itk_component(canvas) configure \
	-scrollregion "0 0 $display_size_x $display_size_y" 

    if {$zoom < 64} {
	    zoomluke "in"
	    zoomluke "out"
	    plotSpots
    } else {
	    zoomluke "out"
	    zoomluke "in"
	    plotSpots
    }
    updateView
    plotSpots
    $::session updatePredictions

}




body ImageDisplay::resetViewerSize { { a_zoom "" } } {
#puts "RESETVIEWERSIZE"

    if {$a_zoom != ""} {
	set min_zoom $a_zoom
    }

    $itk_component(canvas) delete marker

    # Initialise position records, and previous positoin records for use in
    #  marker positioning
    set left_x 1
    set top_y 1
    set old_left_x 1
    set old_top_y 1
    set right_x $detector_size_x
    set bottom_y $detector_size_y
    set zoom $min_zoom
    set old_zoom $zoom

    # Calculate desired canvas size (To fit image at 1:4 resolution)
    set display_size_x [expr $detector_size_x * $min_zoom]
    set display_size_y [expr $detector_size_y * $min_zoom]


    # Set the canvas size
    $itk_component(canvas) configure \
	-width $display_size_x \
	-height $display_size_y
	
    update

    set cnvsfrm_width [winfo width $itk_component(main)]
    set cnvsfrm_height [winfo height $itk_component(main)]	

    if {$display_size_x > $cnvsfrm_width} {
	    set display_size_x $cnvsfrm_width
    }
    
    if {$display_size_y > $cnvsfrm_height} {
	    set display_size_y $cnvsfrm_height
    }

    $itk_component(canvas) configure \
	-width $display_size_x \
	-height $display_size_y

    # Redisplay spots
    plotSpots

    # Redisplay predictions
    $::session updatePredictions

    # Refresh view
    updateView

    $itk_component(canvas) configure \
	    -scrollregion "0 0 $display_size_x $display_size_y" 

    zoomluke "in"
    zoomluke "out"
    plotSpots
#	puts $min_zoom
}


##############################################################################
# Method to refresh the view in the viewer
##############################################################################

body ImageDisplay::updateView { { update_markers "1" } } {
    # hide magnifier, setting flag to indicate that it should be reshown at end
    if {[wm state $itk_component(magnifier)] == "normal"} {
	set mag_shown 1
	wm withdraw $itk_component(magnifier)
    } else {
	set mag_shown 0
    }
    
    # Remove zoom rectangle from the canvas
    $itk_component(canvas) delete zoom_box

    # Remove pick's raw data
    $itk_component(canvas) delete pick

    # Mag exp 18/10/04
    #$itk_component(mag_canvas) delete zoom_image

    # Increment right and bottom margins by 1, unless at limit
    if {$right_x < $detector_size_x} {
	set l_uber_right_x [expr $right_x + 1]
    } else {
	set l_uber_right_x $right_x
    }
    if {$bottom_y < $detector_size_y} {
	set l_uber_bottom_y [expr $bottom_y + 1]
    } else {
	set l_uber_bottom_y $bottom_y
    }

    # Create new image for main view
    if {($::tcl_platform(os) == "Linux") && ([$::session getParameterValue detector_manufacturer] == "ADSC") && ($::env(TKIMAGELOAD) == "1")} {
	set l_x1 $top_y
	set l_x2 $l_uber_bottom_y
	set l_y1 $left_x
	set l_y2 $l_uber_right_x

    } else {
	set l_x1 $left_x
	set l_x2 $l_uber_right_x
	set l_y1 $top_y
	set l_y2 $l_uber_bottom_y
    }
 
   if {$zoom < 1} {
	$view_image copy $master_image \
	    -from $l_x1 $l_y1 $l_x2 $l_y2 \
	    -subsample [expr int(1 / $zoom)] [expr int(1 / $zoom)]
    } else {
	$view_image copy $master_image \
	    -from $l_x1 $l_y1 $l_x2 $l_y2 \
	    -zoom [expr int($zoom)]
    }
    if {($::tcl_platform(os) == "Linux") && ([$::session getParameterValue detector_manufacturer] == "ADSC") && ($::env(TKIMAGELOAD) == "1")} {
	imageTranspose $view_image
    }
    if {$::first_view_update} {
	set ::timer004 [clock clicks -milli]
	set ::first_view_update 0
    }

    # If at max zoom, send pick command
    if {$zoom == $max_zoom} {
 	$::mosflm pick $image_object \
 	   [expr $left_x + 10] [expr $top_y + 10] 21 21
    }

    # Repositions markers ####################

    if {$update_markers} {

	# Move markers back so their origin is at 0,0, so that scaling can be done
	# relative to 0,0.
	$itk_component(canvas) move marker [expr ($old_zoom * ($old_left_x - 1)) - ($old_zoom / 2) * ($old_zoom > 1)] [expr ($old_zoom * ($old_top_y - 1)) - ($old_zoom / 2) * ($old_zoom > 1)]
	
	# Scale markers to match current zoom
	$itk_component(canvas) scale marker 0 0 [expr $zoom / $old_zoom] [expr $zoom / $old_zoom]
	
	# Move spots to match current view position
	$itk_component(canvas) move marker [expr (-$zoom * ($left_x - 1)) + ($zoom / 2) * ($zoom > 1)] [expr (-$zoom * ($top_y - 1)) + ($zoom / 2) * ($zoom > 1)]
    }
    # #############

    # Raise spots to make them viewable, if their toolbutton is pushed in
    if {[$itk_component(spots) query]} {
	$itk_component(canvas) raise crosses view_image
    } else {
	$itk_component(canvas) lower crosses view_image
    }

    if {[$itk_component(predictions) query] == "active"} {
	$itk_component(canvas) raise predictions
    }

	plotHKL


    # Update overlays with new view information
    Marking::updateViewInformation $left_x $top_y $zoom $pixel_size $display_size_x $display_size_y  $detector_size_x $detector_size_y

    # Plot beam if necessary
    if {[$itk_component(beam) query]} {
	Overlay::plotParameter beam_x
    } else {
	Overlay::clearParameter beam_x
    }

    # Plot backstop if necessary
    if {[$itk_component(backstop) query]} {
	#plotBackstop
    } else {
    }

    # Plot masks if necessary
    if {[$itk_component(masks) query]} {
	Overlay::plotParameter backstop_radius
	Overlay::plotParameter backstop_x
	Mask::plotAll
    } else {
	Overlay::clearParameter backstop_x
	Overlay::clearParameter backstop_radius
	Mask::clearAll
    }

    # Plot spot search boundaries if necessary
    if {[$itk_component(spotfinding_settings) query]} {
	Overlay::plotParameter search_area_min_radius
	Overlay::plotParameter search_area_max_radius
	Overlay::plotParameter bbox_orientation
    } else {
	Overlay::clearParameter search_area_min_radius
	Overlay::clearParameter search_area_max_radius
	Overlay::clearParameter bbox_orientation
    }

    # Plot resolution limits if necessary
    if {[$itk_component(resolution_limits) query]} {
	Overlay::plotParameter high_resolution_limit
	Overlay::plotParameter low_resolution_limit
    } else {
	Overlay::clearParameter high_resolution_limit
	Overlay::clearParameter low_resolution_limit
    }
   
    # Plot circlefitting if necessary
    if {[$itk_component(circle_fitting) query]} {
	CircleFit::replot
    }

    # Timing info
    if {[info exists ::stopwatch1]} {
	set ::stopwatch3 [clock clicks -milli]
    } else {
    }

    set t0 [clock clicks -milli]	


    # Display the magnification image in the magnifier canvas
    
    set mag_zoom [expr int($zoom * 4)]
    set mag_origin [$itk_component(mag_canvas) coords "origin"]
    if {[llength $mag_origin] == 2} {
	$itk_component(mag_canvas) scale marker [lindex $mag_origin 0] [lindex $mag_origin 1] [expr $zoom / $old_zoom] [expr $zoom / $old_zoom]
    }
    
    if {$mag_shown} {
	wm deiconify $itk_component(magnifier)
	#raise $itk_component(magnifier) .image
	raise $itk_component(magnifier)
    }

    #update
    
    # Fix zoom bindings and buttons #########################

    set old_zoom $zoom

    # zoom tool
    if {[$itk_component(zoom_tool) query]} {
	setupZoomToolBindings
    }

    # scroll zooming and toolbutton zooms
    if {$zoom < $max_zoom} {
	bind $itk_component(canvas) <ButtonPress-4> [code $this zoom "in" %x %y]
	bind $itk_component(canvas) <Control-ButtonPress-1> [code $this zoom "in" %x %y]
	$itk_component(zoom_in_tool) configure -state normal
    } else {
	bind $itk_component(canvas) <ButtonPress-4> {}
	bind $itk_component(canvas) <Control-ButtonPress-1> {}
	$itk_component(zoom_in_tool) configure -state disabled
    }
    if {$zoom > $min_zoom} {
	bind $itk_component(canvas) <ButtonPress-5> [code $this zoom "out" %x %y]
	bind $itk_component(canvas) <Shift-ButtonPress-1> [code $this zoom "out" %x %y]
	$itk_component(zoom_out_tool) configure -state normal
    } else {
	bind $itk_component(canvas) <ButtonPress-5> {}
	bind $itk_component(canvas) <Shift-ButtonPress-1> {}
	$itk_component(zoom_out_tool) configure -state disabled
    }
	# added by luke on 13 November 2007.
	# Bug 87. Reset view on mac is middle buttor whereas on other platforms it is the 
	# right mouse button
	# By trial and error I figured out that buttonpress-2 is the right button on the mac
	# so I had to include this if statement to check
	if {[tk windowingsystem] == "aqua"} {
		bind $itk_component(canvas) <ButtonPress-2> [code $this resetViewerSizeLuke]
	} else {
	    bind $itk_component(canvas) <ButtonPress-3> [code $this resetViewerSizeLuke]
#The above line used to be bound to resetView but I changed it to resetViewerSizeLuke
	}
#    $itk_component(canvas) configure -cursor $current_cursor

}

##############################################################################
# Method to get the master pixel that contains a given canvas point
##############################################################################

body ImageDisplay::getMasterPixel { x y { a_multi "multi" } } {

    # Calculated as the distance, scaled by the zoom, from the top left pixel
    #  of the displayed image, plus the position of that top left pixel,
    #  remembering to increment initial values to move from the 0-indexed
    #  canvas to the 1-indexed image, and adding half the reciprocal of the zoom
    #  to account for when 1 pixel representes multiple master pixels.
    
    #      set x [expr $left_x + int( floor( (double([incr x])/$zoom) ) )]
    #      set y [expr $top_y + int( floor( (double([incr y])/$zoom) ) )]
    set old_x $x
    set old_y $y
    if {$a_multi == "Multi"} {
	set x [expr $left_x + int(floor((double($x+1)/$zoom) - (1.0 / ($zoom * 2))))]
	set y [expr $top_y + int(floor((double($y+1)/$zoom) - (1.0 / ($zoom * 2))))]
    } else {
	set x [expr $left_x + int(floor(double($x)/$zoom))]
	set y [expr $top_y + int(floor(double($y)/$zoom))]
    }

    return [list $x $y]
}

##############################################################################
# Method to reset the view to show the full image at 0.25 x original size
##############################################################################

body ImageDisplay::resetView { } {
    $itk_component(zoom_out_tool) configure -state "disabled"
    bind $itk_component(canvas) <ButtonPress-5> {}
    set current_cursor [$itk_component(canvas) cget -cursor]
    $itk_component(canvas) configure -cursor watch
    set old_left_x $left_x
    set old_top_y $top_y
    set left_x 1
    set top_y 1
    set right_x $detector_size_x
    set bottom_y $detector_size_y
    set zoom $min_zoom
    updateView
}


#added by luke
body ImageDisplay::fillWindow { a_zoom } {
#    $itk_component(zoom_out_tool) configure -state "disabled"
    bind $itk_component(canvas) <ButtonPress-5> {}
    set current_cursor [$itk_component(canvas) cget -cursor]
    $itk_component(canvas) configure -cursor watch
    set old_left_x $left_x
    set old_top_y $top_y
    set left_x 1
    set top_y 1
    set right_x $detector_size_x
    set bottom_y $detector_size_y

	set zoom $a_zoom
#    updateView
}
################################

##############################################################################
# Click zoom methods
##############################################################################

body ImageDisplay::zoomluke { in_or_out_or_zoom { x "" } { y "" }} {

    if {($x == "") || ($y == "")} {
	set x [expr int($display_size_x / 2)]
	set y [expr int($display_size_y / 2)]
    }

    set num_zoom_clicks 1

    # Calculate new zoom level
    if {$in_or_out_or_zoom == "in"} {
	set new_zoom [expr $zoom * 2]
    } elseif {$in_or_out_or_zoom == "out"} {
	set new_zoom [expr $zoom * 0.5]
    } else {
	set new_zoom $in_or_out_or_zoom
    }

    # Limit further zooming if at max (64x)
    if {$new_zoom >= $max_zoom} {
	set new_zoom $max_zoom
	# unbind the zoom
	bind $itk_component(canvas) <ButtonPress-4> {}
	# NB still need a binding (just one that does nothing)!
	bind $itk_component(canvas) <Control-ButtonPress-1> [code $this noZoom]
	$itk_component(zoom_in_tool) configure -state "disabled"
    } else {
	$itk_component(zoom_in_tool) configure -state "normal"
    }
    if {$new_zoom <= $min_zoom} {
	set new_zoom $min_zoom
	# unbind the zoom
	bind $itk_component(canvas) <ButtonPress-5> {}
	# NB still need a binding (just one that does nothing)!
	bind $itk_component(canvas) <Shift-ButtonPress-1> [code $this noZoom]
	$itk_component(zoom_out_tool) configure -state "disabled"
    } else {
	$itk_component(zoom_out_tool) configure -state "normal"
    }

    #Calculate the centre master pixel
    foreach { pix_c_x pix_c_y } [getMasterPixel [expr int($display_size_x / 2)] [expr int($display_size_y / 2)]] break

    #Calculate which master pixel was clicked on
    foreach { pix_x pix_y } [getMasterPixel $x $y] { }

    # Calculate new centre pixel
    set pix_x [expr int($pix_x + (($pix_c_x - $pix_x) * (double($zoom) / $new_zoom)))]
    set pix_y [expr int($pix_y + (($pix_c_y - $pix_y) * (double($zoom) / $new_zoom)))]


   # Calculate number of pixels in new view
    set new_num_pix_x [expr double($display_size_x) / $new_zoom]
    set new_num_pix_y [expr double($display_size_y) / $new_zoom]

    # Calculate index of new top-left pixel
    set new_left_x [expr $pix_x - round(double($new_num_pix_x-1)/2)]
    set new_top_y [expr $pix_y - round(double($new_num_pix_y-1)/2)]

    # Calculate index of new bottom-right pixel
    set new_right_x [expr $new_left_x + int(ceil($new_num_pix_x))]
    set new_bottom_y [expr $new_top_y + int(ceil($new_num_pix_y))]

    # If the position of the new top-left pixel means the new zoom
    #  area extends outside the detector, change it accordingly
    if {$new_left_x < 1} {
	set new_left_x 1
	set new_right_x [expr int(ceil($new_num_pix_x))]
    } elseif {$new_right_x > $detector_size_x} {
	set new_right_x $detector_size_x
	set new_left_x [expr $detector_size_x - int(ceil($new_num_pix_x)) + 1]
    }
    if {$new_top_y < 1} {
	set new_top_y 1
	set new_bottom_y [expr int(ceil($new_num_pix_y))]
    } elseif {$new_bottom_y > $detector_size_y} {
	set new_bottom_y $detector_size_y
	set new_top_y [expr $detector_size_y - int(ceil($new_num_pix_y)) + 1]
    }
    zoomViewLuke $new_left_x $new_top_y $new_right_x $new_bottom_y $new_zoom

#    $itk_component(canvas) configure -cursor $current_cursor_remember
#	puts $current_cursor_remember
#	set current_cursor $current_cursor_remember

}

body ImageDisplay::zoom { in_or_out_or_zoom { x "" } { y "" } } {
	set current_cursor_remember [$itk_component(canvas) cget -cursor]
	
    if {($x == "") || ($y == "")} {
	set x [expr int($display_size_x / 2)]
	set y [expr int($display_size_y / 2)]
    }

    set num_zoom_clicks 1


    # Calculate new zoom level
    if {$in_or_out_or_zoom == "in"} {
	set new_zoom [expr $zoom * 2]
    } elseif {$in_or_out_or_zoom == "out"} {
	set new_zoom [expr $zoom * 0.5]
    } else {
	set new_zoom $in_or_out_or_zoom
    }

    # Limit further zooming if at max (64x)
    if {$new_zoom >= $max_zoom} {
	set new_zoom $max_zoom

	# unbind the zoom
	bind $itk_component(canvas) <ButtonPress-4> {}
	# NB still need a binding (just one that does nothing)!
	bind $itk_component(canvas) <Control-ButtonPress-1> [code $this noZoom]
	$itk_component(zoom_in_tool) configure -state "disabled"
    } else {
	$itk_component(zoom_in_tool) configure -state "normal"
    }
    if {$new_zoom <= $min_zoom} {
	set new_zoom $min_zoom
	# unbind the zoom
	bind $itk_component(canvas) <ButtonPress-5> {}
	# NB still need a binding (just one that does nothing)!
	bind $itk_component(canvas) <Shift-ButtonPress-1> [code $this noZoom]
	$itk_component(zoom_out_tool) configure -state "disabled"
    } else {
	$itk_component(zoom_out_tool) configure -state "normal"
    }
    #Calculate the centre master pixel
    foreach { pix_c_x pix_c_y } [getMasterPixel [expr int($display_size_x / 2)] [expr int($display_size_y / 2)]] break


    #Calculate which master pixel was clicked on
    foreach { pix_x pix_y } [getMasterPixel $x $y] { }

    # Calculate new centre pixel
    set pix_x [expr int($pix_x + (($pix_c_x - $pix_x) * (double($zoom) / $new_zoom)))]
    set pix_y [expr int($pix_y + (($pix_c_y - $pix_y) * (double($zoom) / $new_zoom)))]

#added by luke 7
#	puts $display_size_x
	set display_size_x [winfo width $itk_component(main)]
	set display_size_y [winfo height $itk_component(main)]
#
#   $itk_component(canvas) configure \
#	-width $display_size_x \
#	-height $display_size_y
#	
#	updateView
#######################

   # Calculate number of pixels in new view
    set new_num_pix_x [expr double($display_size_x) / $new_zoom]
    set new_num_pix_y [expr double($display_size_y) / $new_zoom]
#added by luke 
###########################################
    # Calculate index of new top-left pixel
    set new_left_x [expr $pix_x - round(double($new_num_pix_x-1)/2)]
    set new_top_y [expr $pix_y - round(double($new_num_pix_y-1)/2)]

    # Calculate index of new bottom-right pixel
    set new_right_x [expr $new_left_x + int(ceil($new_num_pix_x))]
    set new_bottom_y [expr $new_top_y + int(ceil($new_num_pix_y))]


    # If the position of the new top-left pixel means the new zoom
    #  area extends outside the detector, change it accordingly
    if {$new_left_x < 1} {
	set new_left_x 1
	set new_right_x [expr int(ceil($new_num_pix_x))]
    } elseif {$new_right_x > $detector_size_x} {
	set new_right_x $detector_size_x
	set new_left_x [expr $detector_size_x - int(ceil($new_num_pix_x)) + 1]
    }
    if {$new_top_y < 1} {
	set new_top_y 1
	set new_bottom_y [expr int(ceil($new_num_pix_y))]
    } elseif {$new_bottom_y > $detector_size_y} {
	set new_bottom_y $detector_size_y
	set new_top_y [expr $detector_size_y - int(ceil($new_num_pix_y)) + 1]
    }
    zoomView $new_left_x $new_top_y $new_right_x $new_bottom_y $new_zoom

    $itk_component(canvas) configure -cursor $current_cursor_remember
	set current_cursor $current_cursor_remember


}


body ImageDisplay::zoomViewLuke { x1 y1 x2 y2 new_zoom} {

#    set current_cursor [$itk_component(canvas) cget -cursor]
#    $itk_component(canvas) configure -cursor watch
    set zoom $new_zoom
    set num_zoom_clicks 0
    set old_left_x $left_x
    set old_top_y $top_y
    set left_x $x1
    set top_y $y1
    set right_x $x2
    set bottom_y $y2
#    updateView
}


body ImageDisplay::zoomView { x1 y1 x2 y2 new_zoom} {
    set current_cursor [$itk_component(canvas) cget -cursor]
    $itk_component(canvas) configure -cursor watch
    set zoom $new_zoom
    set num_zoom_clicks 0
    set old_left_x $left_x
    set old_top_y $top_y
    set left_x $x1
    set top_y $y1
    set right_x $x2
    set bottom_y $y2
#added by luke
	resizeWindow
###############
    updateView
#added by luke
	plotSpots
################
}


body ImageDisplay::zoomClick { a_x a_y } {
    storePreBindings "zoom"
    # Store coordinates of initial zoom click
    set zoom_click_x [getCoorx $a_x]
    set zoom_click_y [getCoory $a_y]
	
#	set zoom_click_x [$itk_component(canvas) canvasx $a_x]
#	set zoom_click_y [$itk_component(canvas) canvasy $a_y]
 
   # Create the zoom rectangles (clicked rectangle and prospective view)
    $itk_component(canvas) create rectangle \
	$zoom_click_x $zoom_click_y $zoom_click_x $zoom_click_y \
	-tags zoom_click_rectangle \
	-outline gold \
	-dash { 1 3 }
    $itk_component(canvas) create rectangle \
	$zoom_click_x $zoom_click_y $zoom_click_x $zoom_click_y \
	-tags zoom_view_rectangle \
	-outline gold
    # Setup bindings for drag and release    
    bind $itk_component(canvas) <Motion> [code $this zoomDrag %x %y]
    bind $itk_component(canvas) <ButtonRelease-1> [code $this zoomRelease %x %y]
}

body ImageDisplay::zoomDrag { a_x a_y } {
    $itk_component(canvas) coords zoom_click_rectangle \
	$zoom_click_x $zoom_click_y $a_x $a_y
    if {($a_x != $zoom_click_x) || ($a_y != $zoom_click_y)} {
	# Calculate max zoom that will fit select rectangle
	set l_width [expr $a_x - $zoom_click_x]
	set l_height [expr $a_y - $zoom_click_y]
	set l_centre_x [expr $zoom_click_x + ($l_width / 2)]
	set l_centre_y [expr $zoom_click_y + ($l_height / 2)]
	set l_max_dim [expr abs($l_width) > abs($l_height) ? $l_width : $l_height]
	set l_num_pix_to_show [expr abs($l_max_dim) / $zoom]
	if {$l_num_pix_to_show == 0} {
	    set l_num_pix_to_show 1
	}
	set l_min_display_dim [expr $display_size_x < $display_size_y ? $display_size_x : $display_size_y]
	set l_new_zoom [expr $l_min_display_dim / $l_num_pix_to_show]
	set l_old_new_zoom $l_new_zoom
	set l_new_zoom [expr pow(2,int(floor(log($l_new_zoom) / log(2))))]
	if {$l_new_zoom > $max_zoom} {
	    set l_new_zoom $max_zoom
	} elseif {$l_new_zoom < $min_zoom} {
	    set l_new_zoom $min_zoom
	}
	# Calculate draw size of new rectangle
	set l_real_width [expr int($display_size_x * ($zoom / $l_new_zoom))]
	set l_real_height [expr int($display_size_y * ($zoom / $l_new_zoom))]
	# Calculate new corner positions
	set l_new_x1 [expr $l_centre_x - ($l_real_width / 2)]
	set l_new_x2 [expr $l_new_x1 + $l_real_width - 1]
	set l_new_y1 [expr $l_centre_y - ($l_real_height / 2)]
	set l_new_y2 [expr $l_new_y1 + $l_real_height - 1]
	if {$l_new_x1 < 0} {
	    set l_new_x1 0
	    set l_new_x2 [expr $l_real_width - 1]
	} elseif {$l_new_x2 >= $display_size_x} {
	    set l_new_x2 [expr $display_size_x - 1]
	    set l_new_x1 [expr $display_size_x - $l_real_width]
	}
	if {$l_new_y1 < 0} {
	    set l_new_y1 0
	    set l_new_y2 [expr $l_real_height - 1]
	} elseif {$l_new_y2 >= $display_size_y} {
	    set l_new_y2 [expr $display_size_y - 1]
	    set l_new_y1 [expr $display_size_y - $l_real_height]
	}
	# Draw real view rectangle
	$itk_component(canvas) coords zoom_view_rectangle \
	    $l_new_x1 $l_new_y1 $l_new_x2 $l_new_y2
	# Return upper left corner and new zoom
	set l_return [eval list [getMasterPixel $l_new_x1 $l_new_y1 "not_multi"] $l_new_zoom]
	foreach { l_mpx l_mpy } $l_return break
	return $l_return 
    } else {
	return {}
    }
}

body ImageDisplay::zoomRelease { a_x a_y } {
    $itk_component(canvas) delete zoom_click_rectangle zoom_view_rectangle
    set l_zoom ""
    foreach { l_x1 l_y1 l_zoom } [zoomDrag $a_x $a_y] break
	if {$l_zoom != ""} {
	    set l_x2 [expr int(ceil($l_x1 + ($display_size_x / $l_zoom) - 1))]
	    set l_y2 [expr int(ceil($l_y1 + ($display_size_y / $l_zoom) - 1))]
	    if {$l_x2 > $detector_size_x} {
		set l_x1 [expr $l_x1 - ($l_x2 - $detector_size_x)]
		set l_x2 $detector_size_x
	    }
	    if {$l_y2 > $detector_size_y} {
		set l_y1 [expr $l_y1 - ($l_y2 - $detector_size_y)]
		set l_y2 $detector_size_y
	    }
	    zoomView $l_x1 $l_y1 $l_x2 $l_y2 $l_zoom
	}
    # Setup bindings for drag and release
    bind $itk_component(canvas) <Motion> {}
    bind $itk_component(canvas) <ButtonRelease-1> {}
    restorePreBindings "zoom"
}

body ImageDisplay::toggleSpots { state } {
    if {$state == "1"} {
	$itk_component(canvas) raise crosses view_image
	$itk_component(mag_canvas) raise crosses zoom_image
    } else {
	$itk_component(canvas) lower crosses view_image
	$itk_component(mag_canvas) lower crosses zoom_image
	$itk_component(spotfinding) cancel
	$itk_component(spotfinding) cancel
	$itk_component(spotdeleting) cancel
    }
}

body ImageDisplay::togglePredictions { state } {
   if {$state == "1"} {
	$itk_component(canvas) raise prediction
	$itk_component(mag_canvas) raise prediction
   } else {
	$itk_component(canvas) lower prediction
	$itk_component(mag_canvas) lower prediction
   }
}

body ImageDisplay::magnifyStart { a_type x y } {
    storePreBindings "magnify"
    if {$a_type == "ButtonPress-1"} {
	# Setup bindings to cancel magnification on non-shift motion or Leave
	bind $itk_component(canvas) <Motion> [code $this magnifyMove %x %y 13 13]
	bind $itk_component(canvas) <ButtonRelease-1> [code $this magnifyEnd] 
	bind $itk_component(canvas) <Leave> [code $this magnifyEnd] 
    } else {
	bind $itk_component(canvas) <Shift-Motion> [code $this magnifyMove %x %y 13 13]
	bind $itk_component(canvas) <Motion> [code $this magnifyEnd] 
	bind $itk_component(canvas) <Leave> [code $this magnifyEnd] 
    }
    # N.B. magnifyMove called twice, as single call inexplicably fails to work!
    magnifyMove $x $y 13 13
    magnifyMove $x $y 13 13
    # SHow magnifier window if necessary
    wm deiconify $itk_component(magnifier)
    raise $itk_component(magnifier)
    # Change cursor
    if {[tk windowingsystem] == "x11"} {
	$itk_component(canvas) configure -cursor $::cursor::box
	set current_cursor $::cursor::box
    } else {
	$itk_component(canvas) configure -cursor cross
	set current_cursor cross
    }	    
}

body ImageDisplay::magnifyMove { x y dx dy} {
    # Move magnifier position
    wm geometry $itk_component(magnifier) +[expr [winfo pointerx $itk_component(magnifier)] + $dx]\+[expr [winfo pointery $itk_component(magnifier)] + $dy]
    # raise $itk_component(magnifier) $itk_component(canvas)
    # raise $itk_component(magnifier)

    # Calculate image pixel mouse is over
    foreach { xi yi } [getMasterPixel $x $y] { }

    # Calculate area to be shown
    set left_border [expr int($xi - (10 / $zoom))]
    set top_border [expr int($yi - (10 / $zoom))]
    set right_border [expr  int($xi + (10 / $zoom) + 1)]
    set bottom_border [expr int($yi + (10 / $zoom) + 1)]

    if {($::tcl_platform(os) == "Linux") && ([$::session getParameterValue detector_manufacturer] == "ADSC") && ($::env(TKIMAGELOAD) == "1")} {
	set im_left_border [expr int($yi - (10 / $zoom))]
	set im_top_border [expr int($xi - (10 / $zoom))]
	set im_right_border [expr int($yi + (10 / $zoom) + 1)]
	set im_bottom_border [expr  int($xi + (10 / $zoom) + 1)]
    } else {
	set im_left_border [expr int($xi - (10 / $zoom))]
	set im_top_border [expr int($yi - (10 / $zoom))]
	set im_right_border [expr  int($xi + (10 / $zoom) + 1)]
	set im_bottom_border [expr int($yi + (10 / $zoom) + 1)]

    }

    if {$left_border < 0 || $top_border < 0 || $right_border > $detector_size_x || $bottom_border > $detector_size_y} {
	return
    }

    # Update zoom image to reflect new position
    if {($zoom * 4) >= 1} {
	$zoom_image copy $master_image \
	    -from $im_left_border $im_top_border $im_right_border $im_bottom_border \
	    -zoom [expr int($zoom * 4)]
    } else {
	$zoom_image copy $master_image \
	    -from $im_left_border $im_top_border $im_right_border $im_bottom_border \
	    -subsample [expr int(1 /($zoom*4))] [expr int(1 /($zoom*4))] 
    }

    # Transpose if necessary
    if {($::tcl_platform(os) == "Linux") && ([$::session getParameterValue detector_manufacturer] == "ADSC") && ($::env(TKIMAGELOAD) == "1")} {
	imageTranspose $zoom_image
    }

    # Move markers to new position
    foreach { origin_x origin_y } [$itk_component(mag_canvas) coords "origin"] { }
    # N.B. Shift of 1.5 added to make result match view image... :O
    $itk_component(mag_canvas) move marker [expr -$origin_x - (($left_border-1.5) * ($zoom * 4))] [expr -$origin_y - (($top_border-1.5) * ($zoom * 4))] 

}

body ImageDisplay::magnifyEnd { } {

    bind $itk_component(canvas) <Motion> {}
    bind $itk_component(canvas) <Shift-Motion> {}
    bind $itk_component(canvas) <ButtonRelease-1> {}
    bind $itk_component(canvas) <Leave> {}
    restorePreBindings "magnify"
    $itk_component(canvas) configure -cursor $::cursor::zoom
    wm withdraw $itk_component(magnifier)
}


body ImageDisplay::plotSpots { } {

    # Don't bother if there is no image
    if {$image_object == ""} {
	return
    }

    # Remove exisitng spots
    $itk_component(canvas) delete crosses
    $itk_component(mag_canvas) delete crosses

    # Get any spotlist associated with the current image
    set l_spotlist [$image_object getSpotlist]
    
    # If there is a spotlist...
    if {$l_spotlist != ""} {
		
	# plot spots on each canvas
	$l_spotlist plotSpots $i_sig_i $itk_component(canvas) $itk_component(mag_canvas)
	
	# position the spots
	positionMarkers crosses
	
	# Raise spots to make them viewable, if their toolbutton is pushed in
	if {[$itk_component(spots) query]} {
	    $itk_component(canvas) raise crosses view_image
	} else {
	    $itk_component(canvas) lower crosses view_image
	}
    }    
}

body ImageDisplay::positionMarkers { a_tag } {
 
    # position spots plotted in their 1:1 pixel location
    #  to match current zoom and view position 

    # Scale crosses to match current zoom
    $itk_component(canvas) scale $a_tag 0 0 $zoom $zoom
    $itk_component(mag_canvas) scale $a_tag 0 0 [expr $zoom * 4] [expr $zoom * 4]
	
    # Move spots on view canvas to match current view position
    $itk_component(canvas) move $a_tag [expr (-$zoom * ($left_x - 1)) + ($zoom / 2) * ($zoom > 1)] [expr (-$zoom * ($top_y - 1)) + ($zoom / 2) * ($zoom > 1)]

    # Move spots on mag canvas to match current origin
    eval $itk_component(mag_canvas) move $a_tag  [$itk_component(mag_canvas) coords "origin"]
}
 
body ImageDisplay::parseHistogram { a_dom } {

    # Read the intensity histogram and contrast info
    $itk_component(contrast_palette) updateHistogram \
	[$a_dom selectNodes string(//pixel_intensities)]

    # Set the flag to indicate that there's an image displayed
    set image_is_open "1"

}

#########################################################
#########################################################
# Functions for interactively setting spot finding params
#########################################################
#########################################################

body ImageDisplay::query { var } {
    return [set $var]
}

################################################################################
################################################################################
################################################################################

body ImageDisplay::updateSetting { a_parameter a_value } {
    set $a_parameter $a_value
    if {[$itk_component(backstop) query]} {
	#plotBackstop
    }
    if {[$itk_component(spotfinding_settings) query]} {
	#plotSpotSearchSettings
	#Overlay::plotParameter search_area_min_radius
    }
    if {[$itk_component(resolution_limits) query]} {
	#plotResolutionLimits
    }
    
    #Overlay::updateParameter $a_parameter $a_value

}

body ImageDisplay::toggleSpotfindingParameters { state } {
   if {$state == "1"} {
       if {[$::session forceBeamSetting]} {
	   Overlay::plotParameter search_area_min_radius
	   Overlay::plotParameter search_area_max_radius
	   Overlay::plotParameter bbox_orientation

       } else {
	   $itk_component(spotfinding_settings) cancel
       }
   } else {
       Overlay::clearParameter search_area_min_radius
       Overlay::clearParameter search_area_max_radius
       Overlay::clearParameter bbox_orientation
   }
}

body ImageDisplay::toggleBeam { state } {
#	puts "Show beam state is $state"
   if {$state == "1"} {
       if {[$::session forceBeamSetting]} {
	   Overlay::plotParameter beam_x
       } else {
	   $itk_component(beam) cancel
       }
   } else {
       Overlay::clearParameter beam_x
   }
}

body ImageDisplay::toggleMasks { state } {
   if {$state == "1"} {
       if {[$::session forceBeamSetting]} {
	   Overlay::plotParameter backstop_radius
	   Overlay::plotParameter backstop_x
	   Mask::plotAll
	   #plotBackstop
       } else {
	   $itk_component(backstop) cancel
       }
   } else {
       Overlay::clearParameter backstop_x
       Overlay::clearParameter backstop_radius
       Mask::clearAll
   }
}

body ImageDisplay::toggleResolutionLimits { state } {
   if {$state == "1"} {
       if {[$::session forceBeamSetting]} {
	   Overlay::plotParameter high_resolution_limit
	   Overlay::plotParameter low_resolution_limit
       } else {
	   $itk_component(resolution_limits) cancel
       }
   } else {
       Overlay::clearParameter high_resolution_limit
       Overlay::clearParameter low_resolution_limit
   }
}

body ImageDisplay::toggleActiveMask { state } {
    set detector_model [$::session getParameterValue detector_model]
    #puts $detector_model
    set rigakus {JUPITER SATURN MERCURY A200}
    if { [lsearch $rigakus $detector_model] >= 0 } {
	#puts "$detector_model found in Rigaku list $rigakus"
	set image_object [.image getImage]
	set l_directory [$image_object getDirectory]
	set l_template [$image_object getTemplate]
	set l_image_number [$image_object getNumber]
	$::mosflm sendCommand "directory $l_directory"
	$::mosflm sendCommand "template $l_template"
	$::mosflm sendCommand "image $l_image_number"
	if {$state == "1"} {
	    set hide_active_mask 1
	    $::mosflm sendCommand "noactive"
	} else {
	    set hide_active_mask 0
	    $::mosflm sendCommand "active"
	}
	$::mosflm sendCommand "xgui on"
	$::mosflm sendCommand "go"
	$::mosflm sendCommand "create_grey_image 0 0"
	$::mosflm sendCommand "return"
	# re-display predictions
	if {[$::session predictionPossible]} {
	    getPredictions
	}
    } else {
	#puts "$detector_model not in Rigaku list $rigakus"
    }
}

body ImageDisplay::toggleInfoLabels { state } {
   if {$state == "1"} {
       if {[$::session forceBeamSetting]} {
	   set show_info_labels 1
       } else {
	   set show_info_labels 0
	   $itk_component(info_labels) cancel noexecute
       }
   } else {
       set show_info_labels 0
   }
}

body ImageDisplay::plotBeam { { a_colour "green" } } {
    $itk_component(canvas) delete beam
    foreach {xc_p yc_p} [getCurrentViewCentrePixel "beam"] break
    $itk_component(canvas) create image $xc_p $yc_p \
	-image ::img::beam_${a_colour}7x7 \
	-tags [list beam_param clickable_spotfinding_param beam]
}

body ImageDisplay::plotBackstop { } {
    plotBackstopRadius green
    plotBackstopCentre green
}

body ImageDisplay::plotSpotSearchSettings { } {

    plotMinSearchRadius red
    plotMaxSearchRadius red
    plotExclusionSegment "vertical" red
    plotExclusionSegment "horizontal" red
    #plotBackgroundBox red
}

body ImageDisplay::plotResolutionLimits { } {
    if {($high_resolution_limit != "") && ([string is double $high_resolution_limit])} {
	plotHighResolutionLimit blue
    } else {
	$itk_component(canvas) delete high_resolution_limit
    }
    if {($low_resolution_limit != "") && ([string is double $low_resolution_limit])} {
	plotLowResolutionLimit blue
    } else {
	$itk_component(canvas) delete low_resolution_limit
    }
}

body ImageDisplay::plotBackstopRadius { a_colour { a_fill "" } } {
    if {$a_fill == ""} {
	set a_fill $a_colour
    }
    $itk_component(canvas) delete backstop_radius
    plotCircle "backstop" $backstop_radius \
	-stipple @[file join $::env(MOSFLM_GUI) bitmaps stipple2.xbm] \
	-outline $a_colour \
	-fill $a_fill \
	-tags [list backstop_param backstop_radius backstop_radius_stipple]
    plotCircle "backstop" $backstop_radius \
	-outline $a_colour \
	-tags [list backstop_param clickable_spotfinding_param backstop_radius backstop_radius_ring]
    $itk_component(canvas) raise backstop_centre
}

body ImageDisplay::plotBackstopCentre { a_colour } {
    $itk_component(canvas) delete backstop_centre
    foreach {xc_p yc_p} [getCurrentViewCentrePixel "backstop"] break
    $itk_component(canvas) create image $xc_p $yc_p \
	-image ::img::backstop_centre_${a_colour}7x7 \
	-tags [list backstop_param clickable_spotfinding_param backstop_centre]
}

body ImageDisplay::plotMaxSearchRadius { a_colour { a_fill "" } } {
    if {$a_fill == ""} {
	set a_fill $a_colour
    }
    $itk_component(canvas) delete search_area_max_radius
    stippleOut $search_area_max_radius 1 \
	-stipple @[file join $::env(MOSFLM_GUI) bitmaps stipple1.xbm] \
	-fill $a_fill \
	-tags [list spotfinding_param search_area_max_radius search_area_max_radius_stipple]
    plotCircle "beam" $search_area_max_radius \
	-outline $a_colour \
	-tags [list spotfinding_param clickable_spotfinding_param search_area_max_radius search_area_max_radius_ring]
}

body ImageDisplay::plotMinSearchRadius { a_colour { a_fill "" } } {
    if {$a_fill == ""} {
	set a_fill $a_colour
    }
    $itk_component(canvas) delete search_area_min_radius
    plotCircle "beam" $search_area_min_radius \
	-stipple @[file join $::env(MOSFLM_GUI) bitmaps stipple2.xbm] \
	-outline $a_colour \
	-fill $a_fill \
	-tags [list spotfinding_param search_area_min_radius search_area_min_radius_stipple]
    plotCircle "beam" $search_area_min_radius \
	-outline $a_colour \
	-tags [list spotfinding_param clickable_spotfinding_param search_area_min_radius search_area_min_radius_ring]
}

body ImageDisplay::plotExclusionSegment { a_direction a_colour { a_fill "" } } {
    if {$a_fill == ""} {
	set a_fill $a_colour
    }
    $itk_component(canvas) delete ${a_direction}_exclusion
    set l_width [set exclusion_segment_${a_direction}]
    if {$l_width < 0.00000001} {
	set l_dash { 1 9 1 9 }
    } else {
	set l_dash { 1 }
	plotExclusion $a_direction $l_width \
	    -stipple @[file join $::env(MOSFLM_GUI) bitmaps stipple1.xbm] \
	    -fill $a_fill \
	    -tags [list spotfinding_param ${a_direction}_exclusion ${a_direction}_exclusion_stipple] \
	    -outline {}
    }
    plotExclusion $a_direction $l_width \
	-dash $l_dash \
	-outline $a_colour \
	-tags [list spotfinding_param clickable_spotfinding_param ${a_direction}_exclusion ${a_direction}_exclusion_ring]
}
  
body ImageDisplay::plotHighResolutionLimit { a_colour { a_fill "" } } {
    if {$a_fill == ""} {
	set a_fill $a_colour
    }

    # Calculate radius
    set high_resolution_radius [expr $distance * tan(2 * asin($wavelength / (2 * $high_resolution_limit)))]

    $itk_component(canvas) delete high_resolution_limit
    stippleOut $high_resolution_radius 1 \
	-stipple @[file join $::env(MOSFLM_GUI) bitmaps stipple2.xbm] \
	-fill $a_fill \
	-tags [list resolution_limit high_resolution_limit high_resolution_limit_stipple]
    plotCircle "beam" $high_resolution_radius \
	-outline $a_colour \
	-tags [list resolution_limit clickable_spotfinding_param high_resolution_limit high_resolution_limit_ring]
}

body ImageDisplay::plotLowResolutionLimit { a_colour { a_fill "" } } {
    if {$a_fill == ""} {
	set a_fill $a_colour
    }
    # Calculate radius
    set low_resolution_radius [expr $distance * tan(2 * asin($wavelength / (2 * $low_resolution_limit)))]

    $itk_component(canvas) delete low_resolution_limit
    plotCircle "beam" $low_resolution_radius \
	-stipple @[file join $::env(MOSFLM_GUI) bitmaps stipple1.xbm] \
	-outline $a_colour \
	-fill $a_fill \
	-tags [list resolution_limit low_resolution_limit low_resolution_limit_stipple]
    plotCircle "beam" $low_resolution_radius \
	-outline $a_colour \
	-tags [list resolution_limit clickable_spotfinding_param low_resolution_limit low_resolution_limit_ring]
}
    


body ImageDisplay::plotCircle { a_item a_radius args } {
    foreach {xc_p yc_p} [getCurrentViewCentrePixel "$a_item"] break
    set radius_p [expr ($a_radius / $pixel_size) * $zoom]
    set x1 [expr $xc_p - $radius_p]
    set y1 [expr $yc_p - $radius_p]
    set x2 [expr $xc_p + $radius_p]
    set y2 [expr $yc_p + $radius_p]
    eval $itk_component(canvas) create oval $x1 $y1 $x2 $y2 $args
}

body ImageDisplay::plotExclusion { a_direction a_width args } {
    foreach {xc_p yc_p} [getCurrentViewCentrePixel "beam"] break
    if {$a_direction == "horizontal"} {
	set x1 -1
	set y1 [expr $yc_p - ((double($a_width) / $pixel_size) * $zoom)]
	set x2 [expr $display_size_x + 1]
	set y2 [expr $yc_p + ((double($a_width) / $pixel_size) * $zoom)]
    } elseif {$a_direction == "vertical"} {
	set x1 [expr $xc_p - ((double($a_width) / $pixel_size) * $zoom)]
	set y1 -1
	set x2 [expr $xc_p + ((double($a_width) / $pixel_size) * $zoom)]
	set y2 [expr $display_size_y + 1]
    }
    eval $itk_component(canvas) create rectangle $x1 $y1 $x2 $y2 $args
}

body ImageDisplay::stippleOutSly { a_radius args } {

    after cancel $stipple_queue

    eval stippleOut [expr $a_radius + 1] 20 $args

    set stipple_queue [after 66 [eval code $this stippleOut $a_radius 1 $args]]
}

body ImageDisplay::stippleOut { a_radius a_step args } {

    #$itk_component(canvas) delete search_area_max_radius_stipple

    foreach {x0 y0} [getCurrentViewCentrePixel "beam"] break
    set r [expr ($a_radius / $pixel_size) * $zoom]
    set coords {}
    set ix [expr $x0 - $r]
    if {$ix < 0} {
	set ix 0
    }
    while {($ix < ($x0 + $r)) && ($ix <= $display_size_y)} {
	set t_side_squared [expr pow($r,2)-pow($ix-$x0,2)]
	if {$t_side_squared < 0} {
	    set t_side_squared 0
	}
	set iy_p [expr $y0 + sqrt($t_side_squared)]
	set iy_n [expr $y0 - sqrt($t_side_squared)]
	if {$iy_p < 0 } {
	    set iy_p 0 
	} elseif {$iy_p > $display_size_y } {
	    set iy_p $display_size_y
	}
	if {$iy_n < 0 } {
	    set iy_n 0
	} elseif {$iy_n > $display_size_y } {
	    set iy_n $display_size_y
	}
	set coords [linsert $coords 0 $ix]
	set coords [linsert $coords 1 $iy_p]
	set coords [linsert $coords end $ix]
	set coords [linsert $coords end $iy_n]
	set ix [expr $ix + $a_step]
    }

    lappend coords [expr $x0 + $r] 
    lappend coords $y0
    lappend coords $display_size_x 
    lappend coords $y0
    lappend coords $display_size_x
    lappend coords 0
    lappend coords 0
    lappend coords 0
    lappend coords 0
    lappend coords $display_size_y
    lappend coords $display_size_x
    lappend coords $display_size_y
    lappend coords $display_size_x
    lappend coords $y0
    lappend coords [expr $x0 + $r] 
    lappend coords $y0

    eval $itk_component(canvas) create polygon $coords $args
}


body ImageDisplay::getCurrentViewCentrePixel { a_item } {
    set t_x [expr ((([set ${a_item}_x] / $pixel_size) - $left_x) * $zoom) + ($zoom / 2)] 
    set t_y [expr ((([set ${a_item}_y] / $pixel_size) - $top_y) * $zoom) + ($zoom / 2)]
    return [list $t_x $t_y]
}
# Interactions #####################################

body ImageDisplay::editSpotSearchParameterStart { a_x a_y } {
    # Do what???
}

body ImageDisplay::editSpotSearchParameterEnd { a_x a_y } {
    # Do what???
    wm withdraw $itk_component(data_labels)
}

body ImageDisplay::showInfoLabelsScroll { a_x a_y} {
	set scrolledx [getCoorx $a_x]
	set scrolledy [getCoory $a_y]
	showInfoLabels $scrolledx $scrolledy
}


body ImageDisplay::showInfoLabels { a_x a_y } {
    # Show resoluion info 
    set l_coords_mm [Marking::c2mCoords [list $a_x $a_y]]
    set l_resolution [$::session calcResolution $l_coords_mm]
    if {$l_resolution == "infinity"} {
	set l_resolution "\u221e"
    }
    $itk_component(data_label_1) configure \
	-text "Resolution: $l_resolution" \
	-fg black \
	-bg white \
	-anchor w
    # Show info for any markers
    set i_label 2
    # Loop through overlapping items- add a small tolerance for when the 
# image is zoomed right out.
    foreach i_item [$itk_component(canvas) find overlapping [expr $a_x - 1] [expr $a_y - 1] [expr $a_x + 1] [expr $a_y + 1]] {
	set l_info_found 0
	# Search for an info class tag, and extract class
	foreach i_tag [$itk_component(canvas) gettags $i_item] {
	    if {[regexp {info_class\((.*)\)} $i_tag l_match l_class]} {
		if {$l_class == "spot"} {
		    if {[$itk_component(spots) query]} {
			set l_info_found 1
		    }
		} elseif {$l_class == "prediction"} {
		    if {[$itk_component(predictions) query]} {
			set l_info_found 1
		    }
		}			    
	    }
	}
	if {$l_info_found} {
	    # Search for an info label tag, and extract label
	    foreach i_tag [$itk_component(canvas) gettags $i_item] {
		if {[regexp {info_label\((.*)\)} $i_tag l_match l_label]} {
		    set l_info_label $l_label
		}
	    }
	    # Search for a info colour tag, and extract colout
	    foreach i_tag [$itk_component(canvas) gettags $i_item] {
		if {[regexp {info_colour\((.*)\)} $i_tag l_match l_colour]} {
		    set l_info_colour $l_colour
		}
	    }
	    # Search for a fg colour tag, and extract colour
	    set l_fg_colour black
	    foreach i_tag [$itk_component(canvas) gettags $i_item] {
		if {[regexp {info_alt_colour\((.*)\)} $i_tag l_match l_colour]} {			    
		    set l_fg_colour $l_colour
		}
	    }
	    $itk_component(data_label_$i_label) configure \
		-text "$l_info_label" \
		-fg $l_fg_colour \
		-bg $l_info_colour \
		-anchor w
	    pack $itk_component(data_label_$i_label) -fill x
	    incr i_label
	}
    }
    # Remove unused labels
    while {$i_label <= 5} {
	pack forget $itk_component(data_label_$i_label)
	incr i_label
    }
    # Show labels by cursor
# must withdraw it first because of bugs with some window managers
    wm withdraw $itk_component(data_labels)
#this is the right one for the labels by the cursor
    set jabberwockx [expr [winfo pointerx $itk_component(canvas)] + 20]
    set jabberwocky [expr [winfo pointery $itk_component(canvas)] + 20]
    wm geometry $itk_component(data_labels) \+$jabberwockx\+$jabberwocky 
    wm deiconify $itk_component(data_labels)
    raise $itk_component(data_labels)
}
body ImageDisplay::editSpotSearchParameterMove { a_x a_y } {
    # Get 'normal colour' for current spot search parameter
    #  (so it can be reset if it's no longer the current one)
    if {[regexp {resolution} $current_spot_search_parameter]} {
	set l_normal_colour blue
    } elseif {[regexp {backstop} $current_spot_search_parameter]} {
	set l_normal_colour green
    } elseif {[string match mask_polygon(*) $current_spot_search_parameter]} {
	set l_normal_colour green
    } else {
	set l_normal_colour red
    }
    # Get closest clickable marker if there is one
    set l_marker [findClosestNearItem $a_x $a_y clickable_spotfinding_param]
    # If there was one...
    if {$l_marker != ""} {
	wm withdraw $itk_component(data_labels)
	# Work out which one it is
	set l_tags [$itk_component(canvas) gettags $l_marker]
	if {[lsearch $l_tags beam] != -1} {
	    set new_spot_search_parameter beam
	} elseif {[lsearch $l_tags backstop_centre] != -1} {
	    set new_spot_search_parameter backstop_centre
	} elseif {[lsearch $l_tags backstop_radius] != -1} {
	    set new_spot_search_parameter backstop_radius
	} elseif {[lsearch $l_tags search_area_max_radius] != -1} {
	    set new_spot_search_parameter search_area_max_radius
	} elseif {[lsearch $l_tags search_area_min_radius] != -1} {
	    set new_spot_search_parameter search_area_min_radius
	} elseif {[lsearch $l_tags vertical_exclusion] != -1} {
	    set new_spot_search_parameter vertical_exclusion
	} elseif {[lsearch $l_tags horizontal_exclusion] != -1} {
	    set new_spot_search_parameter horizontal_exclusion
	} elseif {[lsearch $l_tags high_resolution_limit] != -1} {
	    set new_spot_search_parameter high_resolution_limit
	} elseif {[lsearch $l_tags low_resolution_limit] != -1} {
	    set new_spot_search_parameter low_resolution_limit
	} elseif {[lsearch -glob $l_tags mask_polygon(*)] != -1} {
	    set new_spot_search_parameter [lsearch -inline -glob $l_tags mask_polygon(*)]
	} else {
	    error "Haven't implemented edits of this marker yet!"
	}
	# If it's changed
	if {$new_spot_search_parameter != $current_spot_search_parameter} {
	    # Restore old parameter marker image or colour
	    if {$current_spot_search_parameter == "beam"} {
		$itk_component(canvas) itemconfigure beam \
		    -image ::img::beam_green7x7
	    } elseif {$current_spot_search_parameter == "backstop_centre"} {
		$itk_component(canvas) itemconfigure backstop_centre \
		    -image ::img::backstop_centre_green7x7
	    } else {
		$itk_component(canvas) itemconfigure ${current_spot_search_parameter}_ring \
		    -outline $l_normal_colour
	    }
	    # Colour in active clickable marker (with new image or colour)
	    set l_active_colour gold 
	    if {$new_spot_search_parameter == "beam"} {
		$itk_component(canvas) itemconfigure beam \
		    -image ::img::beam_gold7x7
	    } elseif {$new_spot_search_parameter == "backstop_centre"} {
		$itk_component(canvas) itemconfigure backstop_centre \
		    -image ::img::backstop_centre_gold7x7
	    } elseif {[string match mask_polygon(*) $new_spot_search_parameter]} {
		$itk_component(canvas) itemconfigure $new_spot_search_parameter \
		    -outline $l_active_colour \
		    -fill $l_active_colour 
	    } else {
		$itk_component(canvas) itemconfigure ${new_spot_search_parameter}_ring \
		    -outline $l_active_colour
	    }
	    # update record of current marker
	    set current_spot_search_parameter $new_spot_search_parameter
	}
    } else {
	# Un-colour last current marker, and clear record of current marker
	$itk_component(canvas) itemconfigure ${current_spot_search_parameter}_ring \
	    -outline $l_normal_colour
	$itk_component(canvas) itemconfigure backstop_centre \
		-image ::img::backstop_centre_green7x7
	$itk_component(canvas) itemconfigure beam \
		-image ::img::beam_green7x7
	if {[string match mask_polygon(*) $current_spot_search_parameter]} {
	    $itk_component(canvas) itemconfigure $current_spot_search_parameter \
		-outline $l_normal_colour \
		-fill $l_normal_colour 
	}
	set current_spot_search_parameter ""

	# Show info labels if necessary
	if {$show_info_labels} {
	    set l_info_labels ""
	    # Show resoluion info 
	    foreach {beam_x_p beam_y_p} [getCurrentViewCentrePixel "beam"] break
	    set radius_p [expr sqrt(pow($a_x - $beam_x_p,2)+pow($a_y - $beam_y_p,2))]
	    set radius_mm [expr ($radius_p / $zoom) * $pixel_size]
	    if {[catch {set l_resolution [format %.2f [expr $wavelength / (2 * sin(0.5 * atan($radius_mm / $distance)))]]}]} {
		set l_resolution "\u221e"
	    }

	    $itk_component(data_label_1) configure \
		-text "Resolution: $l_resolution" \
		-fg black \
		-bg white \
		-anchor w
	    # Show info for any markers
	    set i_label 2
	    # Loop through overlaping items
	    foreach i_item [$itk_component(canvas) find overlapping [expr $a_x - 1] [expr $a_y - 1] [expr $a_x + 1] [expr $a_y + 1]] {

		set l_info_found 0
		# Search for an info class tag, and extract class
		foreach i_tag [$itk_component(canvas) gettags $i_item] {
		    if {[regexp {info_class\((.*)\)} $i_tag l_match l_class]} {
			if {$l_class == "spot"} {
			    if {[$itk_component(spots) query]} {
				set l_info_found 1
			    }
			} elseif {$l_class == "prediction"} {
			    if {[$itk_component(predictions) query]} {
				set l_info_found 1
			    }
			}			    
		    }
		}
		if {$l_info_found} {
		    # Search for an info label tag, and extract label
		    foreach i_tag [$itk_component(canvas) gettags $i_item] {
			if {[regexp {info_label\((.*)\)} $i_tag l_match l_label]} {
			    set l_info_label $l_label
			}
		    }
		    # Search for a info colour tag, and extract colout
		    foreach i_tag [$itk_component(canvas) gettags $i_item] {
			if {[regexp {info_colour\((.*)\)} $i_tag l_match l_colour]} {			    set l_info_colour $l_colour
			}
		    }
		    # Search for a fg colour tag, and extract colout
		    set l_fg_colour black
		    foreach i_tag [$itk_component(canvas) gettags $i_item] {
			if {[regexp {info_alt_colour\((.*)\)} $i_tag l_match l_colour]} {			    
			    set l_fg_colour $l_colour
			}
		    }
		    $itk_component(data_label_$i_label) configure \
			-text "$l_info_label" \
			-fg $l_fg_colour \
			-bg $l_info_colour \
			-anchor w
		    pack $itk_component(data_label_$i_label) -fill x
		    incr i_label
		}
	    }
	    # Remove unused labels
	    while {$i_label <= 5} {
		pack forget $itk_component(data_label_$i_label)
		incr i_label
	    }
	    # Show labels by cursor
	    wm deiconify $itk_component(data_labels)
	    raise $itk_component(data_labels)
	    wm geometry $itk_component(data_labels) +[expr [winfo pointerx $itk_component(data_labels)] + 20]\+[expr [winfo pointery $itk_component(data_labels)] + 20]
	}
    }
}

body ImageDisplay::editSpotSearchParameterClick { a_x a_y } {
    # work out if click was "negative" or "positive" for exclusion segments
    foreach {beam_x_p beam_y_p} [getCurrentViewCentrePixel "beam"] break
    if {$a_x < $beam_x_p} {
	set old_vertical_exclusion -1
    } else {
	set old_vertical_exclusion 1
    }
    if {$a_y < $beam_y_p} {
	set old_horizontal_exclusion -1
    } else {
	set old_horizontal_exclusion 1
    }

    # If parameter was clicked on, set up bindings to drag it
    if {$current_spot_search_parameter != ""} {
	bind $itk_component(canvas) <Motion> [code $this editSpotSearchParameterDrag %x %y]
	editSpotSearchParameterDrag $a_x $a_y
    } else {
	# otherwise it must be a zoom click!
	zoomClick $a_x $a_y
    }
}

body ImageDisplay::editSpotSearchParameterDrag { a_x a_y } {
    foreach {beam_x_p beam_y_p} [getCurrentViewCentrePixel "beam"] break
    if {$current_spot_search_parameter == "beam"} {
	foreach {l_master_x l_master_y } [getMasterPixel $a_x $a_y] break
	set beam_x [format %.2f [expr $l_master_x * $pixel_size]]
	set beam_y [format %.2f [expr $l_master_y * $pixel_size]]
	$itk_component(beam_x_e) setValue $beam_x
	$itk_component(beam_y_e) setValue $beam_y
	plotBeam gold
    } elseif {$current_spot_search_parameter == "backstop_centre"} {
	foreach {l_master_x l_master_y } [getMasterPixel $a_x $a_y] break
	set backstop_x [format %.2f [expr $l_master_x * $pixel_size]]
	set backstop_y [format %.2f [expr $l_master_y * $pixel_size]]
	$itk_component(backstop_x_e) setValue $backstop_x
	$itk_component(backstop_y_e) setValue $backstop_y
	plotBackstopRadius green
	plotBackstopCentre gold
    } elseif {$current_spot_search_parameter == "backstop_radius"} {
	foreach {backstop_x_p backstop_y_p} [getCurrentViewCentrePixel "backstop"] break
	set radius_p [expr sqrt(pow($a_x - $backstop_x_p,2)+pow($a_y - $backstop_y_p,2))]
	set $current_spot_search_parameter [format %.2f [expr ($radius_p / $zoom) * $pixel_size]]
	$itk_component(backstop_r_e) setValue $backstop_radius
	plotBackstopRadius gold green
    } elseif {$current_spot_search_parameter == "search_area_max_radius"} {
	set radius_p [expr sqrt(pow($a_x - $beam_x_p,2)+pow($a_y - $beam_y_p,2))]
	set $current_spot_search_parameter [format %.2f [expr ($radius_p / $zoom) * $pixel_size]]
	$itk_component(max_radius_e) setValue $search_area_max_radius
	plotMaxSearchRadius gold red
    } elseif {$current_spot_search_parameter == "search_area_min_radius"} {
	set radius_p [expr sqrt(pow($a_x - $beam_x_p,2)+pow($a_y - $beam_y_p,2))]
	set $current_spot_search_parameter [format %.2f [expr ($radius_p / $zoom) * $pixel_size]]
	$itk_component(min_radius_e) setValue $search_area_min_radius
	plotMinSearchRadius gold red
    } elseif {$current_spot_search_parameter == "vertical_exclusion"} {
	set width_p [expr $a_x - $beam_x_p]
	if {$old_vertical_exclusion < 0} {
	    if {$width_p > 0} {
		set width_p 0
	    }
	    set width_p [expr -1 * $width_p]
	} else {
	    if {$width_p < 0} {
		set width_p 0
	    }
	}
	set exclusion_segment_vertical [format %.2f [expr ($width_p / $zoom) * $pixel_size]]
	$itk_component(exclusion_segment_vertical_e) setValue $exclusion_segment_vertical
	plotExclusionSegment "vertical" gold red
    } elseif {$current_spot_search_parameter == "horizontal_exclusion"} {
	set width_p [expr $a_y - $beam_y_p]
	if {$old_horizontal_exclusion < 0} {
	    if {$width_p > 0} {
		set width_p 0
	    }
	    set width_p [expr -1 * $width_p]
	} else {
	    if {$width_p < 0} {
		set width_p 0
	    }
	}
	set exclusion_segment_horizontal [format %.2f [expr ($width_p / $zoom) * $pixel_size]]
	$itk_component(exclusion_segment_horizontal_e) setValue $exclusion_segment_horizontal
	plotExclusionSegment "horizontal" gold red
    } elseif {$current_spot_search_parameter == "high_resolution_limit"} {
	set radius_p [expr sqrt(pow($a_x - $beam_x_p,2)+pow($a_y - $beam_y_p,2))]
	set radius_mm [expr ($radius_p / $zoom) * $pixel_size]
	set high_resolution_limit [format %.2f [expr $wavelength / (2 * sin(0.5 * atan($radius_mm / $distance)))]]
	$itk_component(max_res_e) setValue $high_resolution_limit
	plotHighResolutionLimit gold blue
    } elseif {$current_spot_search_parameter == "low_resolution_limit"} {
	set radius_p [expr sqrt(pow($a_x - $beam_x_p,2)+pow($a_y - $beam_y_p,2))]
	set radius_mm [expr ($radius_p / $zoom) * $pixel_size]
	set low_resolution_limit [format %.2f [expr $wavelength / (2 * sin(0.5 * atan($radius_mm / $distance)))]]
	$itk_component(min_res_e) setValue $low_resolution_limit
	plotLowResolutionLimit gold blue
    }
}

body ImageDisplay::editSpotSearchParameterRelease { } {
    if {$current_spot_search_parameter == "beam"} {
	$::session updateSetting beam_x $beam_x 1 1 "User" 0
	$::session updateSetting beam_y $beam_y 1 1 "User" 0

	if {[$::session predictionPossible]} {
	    getPredictions
	}
    } elseif {$current_spot_search_parameter == "backstop_centre"} {
	$::session updateSetting backstop_x $backstop_x 1 1 "User" 0
	$::session updateSetting backstop_y $backstop_y 1 1 "User" 0

	if {[$::session predictionPossible]} {
	    getPredictions
	}
    } elseif {$current_spot_search_parameter == "backstop_radius"} {
	$::session updateSetting backstop_radius $backstop_radius 1 1 "User"
	plotBackstopRadius gold green
    } elseif {$current_spot_search_parameter == "search_area_max_radius"} {
	$::session updateSetting search_area_max_radius $search_area_max_radius 1 1 "User"
	plotMaxSearchRadius gold red
    } elseif {$current_spot_search_parameter == "search_area_min_radius"} {
	$::session updateSetting search_area_min_radius $search_area_min_radius 1 1 "User"
	plotMinSearchRadius gold red
    } elseif {$current_spot_search_parameter == "vertical_exclusion"} {
	$::session updateSetting exclusion_segment_vertical $exclusion_segment_vertical 1 1 "User"
	plotExclusionSegment "vertical" gold red
    } elseif {$current_spot_search_parameter == "horizontal_exclusion"} {
	$::session updateSetting exclusion_segment_horizontal $exclusion_segment_horizontal 1 1 "User"
	plotExclusionSegment "horizontal" gold red
    } elseif {$current_spot_search_parameter == "high_resolution_limit"} {
	$::session updateSetting high_resolution_limit $high_resolution_limit 1 1 "User"
	plotHighResolutionLimit gold blue
    } elseif {$current_spot_search_parameter == "low_resolution_limit"} {
	$::session updateSetting low_resolution_limit $low_resolution_limit 1 1 "User"
	plotLowResolutionLimit gold blue
    }
    bind $itk_component(canvas) <Motion> [code $this editSpotSearchParameterMove %x %y]
}
    

body ImageDisplay::infoPointStart { x y } {
    infoPointMove $x $y
    # Show info window
}

body ImageDisplay::infoPointMove { x y } {
    # Calculate pixel position
    # Calculate mm position
    # Calculate resolution
    # search for markers
    # Get info for closest marker
}

body ImageDisplay::infoPointEnd { x y } {
    # Hide info window
}

#########################################################
#########################################################
# Functions for interactive spot finding
#########################################################
#########################################################

body ImageDisplay::toggleZoomTool { a_state } {
    if {$a_state == 1} {
	setupZoomToolBindings
    } else {
	removeZoomToolBindings
    }
}

body ImageDisplay::togglePointerTool { a_state } {
    if {$a_state == 1} {
	Overlay::setupEditBindings
 	bind $itk_component(canvas) <ButtonPress-2> [code $this mousePick %x %y]
    } else {
	Overlay::removeEditBindings
	bind $itk_component(canvas) <ButtonPress-2> ""
    }
}


body ImageDisplay::toggleSpotAddingTool { a_state } {
    if {$a_state == 1} {
	$itk_component(spots) invoke
	setupSpotAddingBindings
    } else {
	removeSpotAddingBindings
    }
}

body ImageDisplay::toggleMaskingTool { a_state } {
    if {$a_state == 1} {
	$itk_component(masks) invoke
	Mask::setupCreationBindings
    } else {
	Mask::removeCreationBindings
    }
}

body ImageDisplay::toggleCircleFittingTool { a_state } {
    if {$a_state == 1} {
	CircleFit::launch $itk_component(canvas)
    } else {
	CircleFit::clear
    }
}

body ImageDisplay::toggleEraserTool { a_state } {
    if {$a_state == 1} {
	setupEraserBindings
    } else {
	removeEraserBindings
    }
}

body ImageDisplay::togglePanningTool { a_state } {
    if {$a_state == 1} {
	setupPanningBindings
    } else {
	removePanningBindings
    }
}

# Global tool bindings (info labels and zoom mod-clicks}

body ImageDisplay::storePreBindings { a_type } {
    array unset pre_${a_type}_bindings *
    foreach i_event [bind $itk_component(canvas)] {
	set pre_${a_type}_bindings($i_event) [bind $itk_component(canvas) $i_event]
	bind $itk_component(canvas) $i_event {}
    }
}

body ImageDisplay::restorePreBindings { a_type } {
    foreach i_event [array names pre_${a_type}_bindings] {
	bind $itk_component(canvas) $i_event [set pre_${a_type}_bindings($i_event)]
    }
}


body ImageDisplay::startInfoLabelsScroll { a_type a_x a_y} {
	set scrolledx [getCoorx $a_x]
	set scrolledy [getCoory $a_y]
	startInfoLabels $a_type $scrolledx $scrolledy
}


body ImageDisplay::startInfoLabels { a_type a_x a_y } {
    # Trigger current leave binding (if there is one), to "mute"
    #  current tool bindings effects
    set l_leave_binding [bind $itk_component(canvas) <Leave>]
    if {$l_leave_binding != ""} {
	uplevel \#0 $l_leave_binding
    }
    # Store cursor for restoring later
    set info_cursor [$itk_component(canvas) cget -cursor]
    # Set question cursor
    $itk_component(canvas) configure -cursor "question_arrow"
    # Store currrent bindings for restoring later
    storePreBindings "info"
    # Show the info label
    showInfoLabels $a_x $a_y
    # Setup bindings for future info labels
    if {$a_type == "Motion"} {
	bind $itk_component(canvas) <Alt-Motion> [code $this showInfoLabelsScroll %x %y]
	bind $itk_component(canvas) <M2-Motion> [code $this showInfoLabels %x %y]
	bind $itk_component(canvas) <Control-Motion> [code $this showInfoLabels %x %y]
	bind $itk_component(canvas) <Motion> [code $this endInfoLabelsScroll %x %y]
	bind $itk_component(canvas) <Leave> [code $this endInfoLabelsScroll %x %y]
    } else {
	bind $itk_component(canvas) <Motion> [code $this showInfoLabelsScroll %x %y]
	bind $itk_component(canvas) <Leave> [code $this endInfoLabelsScroll %x %y]
	bind $itk_component(canvas) <ButtonRelease-1> [code $this endInfoLabelsScroll %x %y]
    }
}

body ImageDisplay::endInfoLabelsScroll { a_x a_y} {
	set scrolledx [getCoorx $a_x]
	set scrolledy [getCoory $a_y]
	endInfoLabels $scrolledx $scrolledy
}


body ImageDisplay::endInfoLabels { a_x a_y } {
    # Restore cursor
    $itk_component(canvas) configure -cursor $info_cursor
    # Hide data labels
    wm withdraw $itk_component(data_labels)
    # Remove info label bindings
    bind $itk_component(canvas) <Alt-Motion> {}
    bind $itk_component(canvas) <M2-Motion> {}
    bind $itk_component(canvas) <Control-Motion> {}
    bind $itk_component(canvas) <Motion> {}
    bind $itk_component(canvas) <Leave> {}
    bind $itk_component(canvas) <ButtonRelease-1> {}
    # Rstore previous bindings
    restorePreBindings "info"
}

# Tool Binding Methods

body ImageDisplay::getCoorx {a_coor} {
#	puts $a_coor
#	puts [$itk_component(canvas) canvasx $a_coor]
	return [$itk_component(canvas) canvasx $a_coor]
	
}

body ImageDisplay::getCoory {a_coor} {
#	puts $a_coor
#	puts [$itk_component(canvas) canvasy $a_coor]
	return [$itk_component(canvas) canvasy $a_coor]
}


body ImageDisplay::setupZoomToolBindings { } {
    $itk_component(canvas) configure -cursor $::cursor::zoom
    bind $itk_component(canvas) <ButtonPress-1> [code $this zoomClick %x %y]
 
    if {[tk windowingsystem] != "aqua"} {
	bind $itk_component(canvas) <Shift-Motion> [code $this magnifyStart Motion %x %y]
    } else {
	bind $itk_component(canvas) <Shift-Motion> [code $this magnifyStart Motion %x  %y]
	#bind $itk_component(canvas) <M1-ButtonPress-1> [code $this magnifyStart "ButtonPress-1" %x %y]
    }
}

body ImageDisplay::removeZoomToolBindings { } {
    $itk_component(canvas) configure -cursor left_ptr
    bind $itk_component(canvas) <ButtonPress-1> {}
    bind $itk_component(canvas) <Button1-Motion> {}
    bind $itk_component(canvas) <Shift-Motion> {}
    bind $itk_component(canvas) <ButtonRelease-1> {}
}

body ImageDisplay::setupSpotAddingBindings { } {   
	$itk_component(canvas) configure -cursor pencil
	bind $itk_component(canvas) <Enter> [code $this placeSpotStart %x %y]
	bind $itk_component(canvas) <ButtonPress-1> [code $this placeSpotClick %x %y]
	bind $itk_component(canvas) <ButtonRelease-1> [code $this placeSpotRelease %x %y]
	bind $itk_component(canvas) <Motion> [code $this placeSpotMove %x %y]
	bind $itk_component(canvas) <Leave> [code $this placeSpotEnd %x %y]
}

body ImageDisplay::removeSpotAddingBindings { } {   
    $itk_component(canvas) configure -cursor left_ptr
    bind $itk_component(canvas) <Enter> {}
    bind $itk_component(canvas) <ButtonPress-1> {}
    bind $itk_component(canvas) <ButtonRelease-1> {}
    bind $itk_component(canvas) <Motion> {}
    bind $itk_component(canvas) <Leave> {}
}

body ImageDisplay::setupEraserBindings { } {
    $itk_component(canvas) configure -cursor $::cursor::eraser
    bind $itk_component(canvas) <Enter> [code $this deleteMarkerStart %x %y]
    bind $itk_component(canvas) <ButtonPress-1> [code $this deleteMarkerClick %x %y]
    bind $itk_component(canvas) <ButtonRelease-1> [code $this deleteMarkerRelease %x %y]
    bind $itk_component(canvas) <Motion> [code $this deleteMarkerMove %x %y]
    bind $itk_component(canvas) <Leave> [code $this deleteMarkerEnd]
}

body ImageDisplay::removeEraserBindings { } {
    $itk_component(canvas) configure -cursor left_ptr
    bind $itk_component(canvas) <Enter> {}
    bind $itk_component(canvas) <ButtonPress-1> {}
    bind $itk_component(canvas) <ButtonRelease-1> {}
    bind $itk_component(canvas) <Motion> {}
    bind $itk_component(canvas) <Leave> {}
}


body ImageDisplay::setupPanningBindings { } {
    $itk_component(canvas) configure -cursor fleur
    bind $itk_component(canvas) <ButtonPress-1> [code $this panningClick %x %y]
    bind $itk_component(canvas) <ButtonRelease-1> [code $this panningRelease]
}

body ImageDisplay::removePanningBindings { } {
    $itk_component(canvas) configure -cursor left_ptr
    bind $itk_component(canvas) <Enter> {}
    bind $itk_component(canvas) <ButtonPress-1> {}
    bind $itk_component(canvas) <ButtonRelease-1> {}
    bind $itk_component(canvas) <Motion> {}
    bind $itk_component(canvas) <Leave> {}
}

body ImageDisplay::panningClick { a_x a_y } {
    # store pixel that was grabbed
    #foreach { grab_x grab_y } [getMasterPixel $a_x $a_y] break
    set grab_x $a_x
    set grab_y $a_y
    # store view position at grab
    set grab_left_x $left_x
    set grab_top_y $top_y
    set grab_right_x $right_x
    set grab_bottom_y $bottom_y
    # setup pan motion binding
    bind $itk_component(canvas) <Motion> [code $this pan %x %y]
}

body ImageDisplay::panningRelease { } {
    bind $itk_component(canvas) <Motion> {}
}

body ImageDisplay::pan { a_x a_y } {
    # get maskter 
    #foreach { l_new_x l_new_y } [getMasterPixel $a_x $a_y] break
    #set l_dx [expr $grab_x - $l_new_x]
    #set l_dy [expr $grab_y - $l_new_y]

    set l_dx [expr int(($grab_x - $a_x) / $zoom)]
    set l_dy [expr int(($grab_y - $a_y) / $zoom)]

    set l_new_left_x [expr $grab_left_x + $l_dx]
    set l_new_top_y [expr $grab_top_y + $l_dy]
    set l_new_right_x [expr $grab_right_x + $l_dx]
    set l_new_bottom_y [expr $grab_bottom_y + $l_dy]

    if {$l_new_left_x < 1} {
	set l_dx [expr 1 - $l_new_left_x ]
	set l_new_left_x [expr $l_new_left_x + $l_dx]
	set l_new_right_x [expr $l_new_right_x + $l_dx]
    } elseif  {$l_new_right_x > $detector_size_x} {
	set l_dx [expr $detector_size_x - $l_new_right_x ]
	set l_new_left_x [expr $l_new_left_x + $l_dx]
	set l_new_right_x [expr $l_new_right_x + $l_dx]
    }
    
    if {$l_new_top_y < 1} {
	set l_dy [expr 1 - $l_new_top_y ]
	set l_new_top_y [expr $l_new_top_y + $l_dy]
	set l_new_bottom_y [expr $l_new_bottom_y + $l_dy]
    } elseif  {$l_new_bottom_y > $detector_size_y} {
	set l_dy [expr $detector_size_y - $l_new_bottom_y ]
	set l_new_top_y [expr $l_new_top_y + $l_dy]
	set l_new_bottom_y [expr $l_new_bottom_y + $l_dy]
    }

    if {($l_new_left_x != $left_x) || ($l_new_top_y != $top_y)} {
	set old_left_x $left_x
	set old_top_y $top_y
	set left_x $l_new_left_x
	set top_y $l_new_top_y
	set right_x $l_new_right_x
	set bottom_y $l_new_bottom_y
	set current_cursor fleur
	updateView
    }
}

body ImageDisplay::toggleTools { tool args } {
    set clear_canvas_bindings 0
    if {$tool != "circle_fitting"} {
	CircleFit::clear
    }
    if {$tool == "magnifiertool"} {
	if {[tk windowingsystem] == "x11"} {
	    $itk_component(canvas) configure -cursor $::cursor::box
	    set current_cursor $::cursor::box
	} else {
	    $itk_component(canvas) configure -cursor cross
	    set current_cursor cross
	}	    
	bind $itk_component(canvas) <Enter> [code $this magnifyStart %x %y]
	bind $itk_component(canvas) <Motion> [code $this magnifyMove %x %y 13 13]
	bind $itk_component(canvas) <ButtonPress-1> {}
	bind $itk_component(canvas) <ButtonRelease-1> {}
	bind $itk_component(canvas) <Leave> [code $this magnifyEnd %X %Y]
	bind $itk_component(mag_canvas) <Leave> [code $this magnifyEnd %X %Y]
    } elseif {$tool == "spotfinding"} {
	$itk_component(spots) invoke
	$itk_component(canvas) configure -cursor pencil
	set current_cursor pencil
	bind $itk_component(canvas) <Enter> [code $this placeSpotStart %x %y]
	bind $itk_component(canvas) <ButtonPress-1> [code $this placeSpotClick %x %y]
	bind $itk_component(canvas) <ButtonRelease-1> [code $this placeSpotRelease %x %y]
	bind $itk_component(canvas) <Motion> [code $this placeSpotMove %x %y]
	bind $itk_component(canvas) <Leave> [code $this placeSpotEnd %x %y]
    } elseif {$tool == "spotdeleting"} {
	$itk_component(spots) invoke
	$itk_component(canvas) configure -cursor hand2
	set current_cursor hand2
	bind $itk_component(canvas) <Enter> [code $this deleteMarkerStart %x %y]
	bind $itk_component(canvas) <ButtonPress-1> [code $this deleteMarkerClick %x %y]
	bind $itk_component(canvas) <ButtonRelease-1> [code $this deleteMarkerRelease %x %y]
	bind $itk_component(canvas) <Motion> [code $this deleteMarkerMove %x %y]
	bind $itk_component(canvas) <Leave> [code $this deleteMarkerEnd]
    } elseif {$tool == "pointer"} {
	#$itk_component(spotfinding_settings) invoke
	$itk_component(canvas) configure -cursor left_ptr
	set current_cursor left_ptr
	Overlay::setupEditBindings
 	bind $itk_component(canvas) <ButtonPress-2> [code $this mousePick %x %y]
    } elseif {$tool == "masking_tool"} {
	$itk_component(masks) invoke
	$itk_component(canvas) configure -cursor left_ptr
	set current_cursor left_ptr
	Mask::setupCreationBindings
    } elseif {$tool == "circle_fitting"} {
	$itk_component(canvas) configure -cursor left_ptr
	set current_cursor left_ptr
	CircleFit::launch $itk_component(canvas)
    } elseif {$tool == "panning_tool"} {
	$itk_component(canvas) configure -cursor fleur
	set current_cursor fleur
	bind $itk_component(canvas) <Enter> {}
	bind $itk_component(canvas) <ButtonPress-1> [code $this grabClick %x %y]
	bind $itk_component(canvas) <ButtonRelease-1> [code $this grabRelease %x %y]
	bind $itk_component(canvas) <Motion> {}
	bind $itk_component(canvas) <Leave> {}
    } elseif {$tool == "masking"} {
	clearCanvasBindings
	Mask::setCreationBindings $itk_component(canvas)
    }
}

body ImageDisplay::clearCanvasBindings { } {
    bind $itk_component(canvas) <Enter> {}
    bind $itk_component(canvas) <ButtonPress-1> {}
    bind $itk_component(canvas) <ButtonRelease-1> {}
    bind $itk_component(canvas) <Motion> {}
    bind $itk_component(canvas) <Leave> {}
}

body ImageDisplay::placeSpotStart { x y } {
    
    placeSpot $x $y
    placeSpot $x $y
    #wm deiconify $itk_component(magnifier)
}

body ImageDisplay::placeSpotMove { x y } {
    after cancel $spot_placing_queue
    set spot_placing_queue [after 5 [code $this placeSpot $x $y]]
    # placeSpot $x $y
}

body ImageDisplay::placeSpot { x y } {

    if {[catch {localSpotSearch $x $y} result]} {
	# Cursor out of canvas area
    } else {
	foreach { spot_x spot_y } $result { }
	
	# Caclulate coords of line endings
	set hix [expr $spot_x - 10]
	set lox [expr $spot_x + 10]
	set hiy [expr $spot_y - 10]
	set loy [expr $spot_y + 10]
	
	# create crosses
	$itk_component(canvas) delete proto_cross
	$itk_component(mag_canvas) delete proto_cross
	$itk_component(canvas) create line $hix $hiy $lox $loy $spot_x $spot_y $lox $hiy $hix $loy -capstyle round -fill yellow -tags proto_cross
	$itk_component(mag_canvas) create line $hix $hiy $lox $loy $spot_x $spot_y $lox $hiy $hix $loy -capstyle round -fill yellow -tags [list proto_cross marker]
	
	positionMarkers proto_cross

    }
    
    # Begin magnification
    magnifyMove $x $y 20 20
}

body ImageDisplay::placeSpotClick { x y } {
    magnifyMove $x $y 20 20
    wm deiconify $itk_component(magnifier)
    raise $itk_component(magnifier)
    $itk_component(spots) invoke
}
    
body ImageDisplay::placeSpotRelease { x y } {

    if {$x > 0 && $y > 0 && $x < [image width $view_image] && $y < [image height $view_image]} {
	# refocus magnification
	magnifyMove $x $y 20 20
	
	# Find the best place to put a spot near the cursor
	foreach { spot_x spot_y } [localSpotSearch $x $y] { }
	
	# Create new spotlist for image, if one doesn't exist
	set l_spotlist [$image_object getSpotlist]
	if {$l_spotlist == ""} {
	    foreach { l_phi_start l_phi_end } [$image_object getPhi] break
	    set l_spotlist [namespace current]::[Spotlist \#auto "empty" $image_object [expr {double($l_phi_start + $l_phi_end) / 2.0}] $pixel_size]
	    $image_object setSpotlist $l_spotlist
	}

	# Add a spot to the spotlist with that location
	set l_new_spot [$l_spotlist addSpotManually $spot_x $spot_y]
	# if successful (i.e. not already got a spot there) draw the spot
	if {$l_new_spot != ""} {
	    $l_new_spot plot $itk_component(canvas)
	    $l_new_spot plot $itk_component(mag_canvas)
	    positionMarkers "cross[$l_new_spot getId]"
	    [.c component indexing] updateSpotFindingResult $image_object
	    # record the spot edit in the history
	    $::session addHistoryEvent "SpotAddEvent" "User action" [$image_object getFullPathName] $l_new_spot
	} else {
	    # Cannot add spot at same location as existing spot
	}
	
    } else {
	# Outside of image
    }

    $itk_component(canvas) delete proto_cross
    $itk_component(mag_canvas) delete proto_cross	
    
    wm withdraw $itk_component(magnifier)
}

body ImageDisplay::placeSpotEnd { x y } {
    
    if {$x >= 0 && \
	    $x <= [winfo width $itk_component(canvas)] && \
	    $y >= 0 && \
	    $y <= [winfo height $itk_component(canvas)]
    } {
 	bind $itk_component(canvas) <Enter> { }
 	#placeSpotMove $x $y
 	#bind $itk_component(canvas) <Enter> [code $this placeSpotStart %x %y]
    } else {
	bind $itk_component(canvas) <Enter> [code $this placeSpotStart %x %y]
	after cancel $spot_placing_queue
	
	$itk_component(canvas) delete proto_cross
	$itk_component(mag_canvas) delete proto_cross	
    }
}

body ImageDisplay::localSpotSearch { x y } {
    # Get pixel clicked on
    foreach { xi yi } [getMasterPixel $x $y] { }
    # Correct for subsampling
    if {$zoom < 1} {
	set xi [expr int($xi - (0.5 * (1.0/$zoom)))] 
	set yi [expr int($yi - (0.5 * (1.0/$zoom)))] 
    }

    # Get neighbourhood data (correcting for 'data -from' 
    #  not including bottom-right column + row

    set left_border [expr (($xi - $search_radius) < 0) ? 0 : ($xi - $search_radius)]
    set top_border [expr (($yi - $search_radius) < 0) ? 0 : ($yi - $search_radius)]
    set right_border [expr ((($xi + $search_radius) + 1) > $detector_size_x) ? $detector_size_x : (($xi + $search_radius) + 1)]
    set bottom_border [expr ((($yi + $search_radius) + 1) > $detector_size_y) ? $detector_size_y : (($yi + $search_radius) + 1)]

    set neighbourhood [$master_image data -from \
			   $left_border \
			   $top_border \
			   $right_border \
			   $bottom_border]

    # Calcluate spot centre

    # initialize row and column arrays
    array unset mrow
    array unset mcol
    set x_count 0
    set y_count 0
    while {$y_count < [llength $neighbourhood]} {
	set mcol($y_count) 0
	incr y_count
    }
    while {$x_count < [llength [lindex $neighbourhood 0]]} {
	set mrow($x_count) 0
	incr x_count
    }

    # add up image neighbourhood in a 'row sums' array and 'col sums' array
    set y_count 0
    foreach row $neighbourhood {
	set x_count 0
	foreach item $row {
	    scan [string range $item 1 end] %x value
	    incr mrow($x_count) $value
	    incr mcol($y_count) $value
	    incr x_count
	}
	incr y_count
    }

    # find the blackest column
    scan "ffffff" %x minval
    set mincol [expr $minval * (($search_radius * 2) + 1)]
    set minrow [expr $minval * (($search_radius * 2) + 1)]
    set minx [expr $search_radius]
    set miny $minx
    foreach name [array names mrow] {
	if {$mrow($name) < $mincol} {
	    set mincol $mrow($name)
	    set minx $name
	}
    }
    # find the blackest row
    foreach name [array names mcol] {
	if {$mcol($name) < $minrow} {
	    set minrow $mcol($name)
	    set miny $name
	}
    }

    set spot_x [expr $left_border + $minx - 1] 
    set spot_y [expr $top_border + $miny - 1] 

    return [list $spot_x $spot_y]
}

########################################################

body ImageDisplay::deleteMarkerStart { x y } {

    #Cursor rubber $itk_component(canvas)

    # N.B. deleteMarkerMove called twice, as single call inexplicably fails to work!
    deleteMarkerMove $x $y
    deleteMarkerMove $x $y
}

body ImageDisplay::deleteMarkerMove { x y } {
    
    #$itk_component(canvas) itemconfigure real_crosses -fill red
    #$itk_component(mag_canvas) itemconfigure real_crosses -fill red

    # Test for masks
    set l_mask [Mask::overMask $x $y]
    if {$l_mask != ""} {
	# unclour any cross
	if {$highlighted_cross != ""} {
	    $itk_component(canvas) itemconfigure real_cross$highlighted_cross -fill $highlighted_cross_colour
	    set highlighted_cross ""
	}
	#  if mask has changed, unhighlight any prev mask and highlight new one
	if {$l_mask != $highlighted_mask} {
	    if {$highlighted_mask != ""} {
		$highlighted_mask unhighlightMask
	    }
	    $l_mask highlightMask
	    set highlighted_mask $l_mask
	} else {
	}
    } else {
	# unhighlight any highlighted mask
	if {$highlighted_mask != ""} {
	    $highlighted_mask unhighlightMask
	    set highlighted_mask ""
	}
	# test for near cross
	set chosen [findClosestNearItem $x $y real_crosses]
	if {$chosen != ""} {
	    # get cross's id
	    set l_tags [$itk_component(canvas) gettags $chosen]
	    foreach i_tag $l_tags {
		if {[regexp {real_cross(\d+)} $i_tag match l_id]} {
		    break
		}
	    }
	    # if cross has changed...
	    if {$l_id != $highlighted_cross} {
		if {$highlighted_cross != ""} {
		    # restore previous cross's colour
		    $itk_component(canvas) itemconfigure real_cross$highlighted_cross -fill $highlighted_cross_colour
		    $itk_component(mag_canvas) itemconfigure real_cross$highlighted_cross -fill $highlighted_cross_colour
		}
		#update record of highlighted cross
		set highlighted_cross $l_id
		# store new cross's colour
		set highlighted_cross_colour [$itk_component(canvas) itemcget real_cross$l_id -fill]
		# colour new cross
		$itk_component(canvas) itemconfigure real_cross$highlighted_cross -fill pink
		$itk_component(mag_canvas) itemconfigure real_cross$highlighted_cross -fill pink
		
		
		$itk_component(canvas) raise $chosen crosses
		$itk_component(mag_canvas) raise $chosen crosses
	    }
	} else {
	    if {$highlighted_cross != ""} {
		# restore previous cross's colour
		$itk_component(canvas) itemconfigure real_cross$highlighted_cross -fill $highlighted_cross_colour
		$itk_component(mag_canvas) itemconfigure real_cross$highlighted_cross -fill $highlighted_cross_colour
		set highlighted_cross ""
	    }    
	}
	magnifyMove $x $y 20 20
    }
}

body ImageDisplay::deleteMarkerEnd { } {
    if {$highlighted_cross != ""} {
	$itk_component(canvas) itemconfigure real_cross$highlighted_cross -fill $highlighted_cross_colour
	$itk_component(mag_canvas) itemconfigure real_cross$highlighted_cross -fill $highlighted_cross_colour
	set highlighted_cross ""
    } elseif {$highlighted_mask != ""} {
	$highlighted_mask unhighlightMask
	set highlighted_mask ""
    }
}

body ImageDisplay::deleteMarkerClick { x y } {

    $itk_component(spots) invoke

    magnifyMove $x $y 20 20
    wm deiconify $itk_component(magnifier)
    raise $itk_component(magnifier)
}

body ImageDisplay::deleteMarkerRelease { x y } {
    if {$highlighted_mask != ""} {
	Mask::deleteMask [$highlighted_mask getNumber]
	set highlighted_mask ""
    } elseif {$highlighted_cross != ""} {
	set l_spotlist [$image_object getSpotlist]
	if {$l_spotlist != ""} {
	    # Find the spot to be deleted
	    set l_spot_to_be_deleted [$l_spotlist getSpotById $highlighted_cross]
	    # add spot edit event to the history
	    $::session addHistoryEvent "SpotDeleteEvent" "User action" [$image_object getFullPathName] $l_spot_to_be_deleted 
	    $l_spotlist deleteSpot "id" $highlighted_cross
	    $itk_component(canvas) delete real_cross$highlighted_cross
	    $itk_component(mag_canvas) delete real_cross$highlighted_cross
	    [.c component indexing] updateSpotFindingResult $image_object
	}
	set highlighted_cross ""
    }
    wm withdraw $itk_component(magnifier)
}

body ImageDisplay::findClosestNearItem { x y tag } {
    set chosen ""
    set l_distance "999"
    set items [$itk_component(canvas) find overlapping \
		   [expr $x - $search_radius] \
		   [expr $y - $search_radius] \
		   [expr $x + $search_radius] \
		   [expr $y + $search_radius]]
    foreach item $items {
	if {[lsearch [$itk_component(canvas) itemcget $item -tags] $tag] != -1} {
	    if {$tag == "real_crosses"} {
		foreach { loc_x loc_y } [lrange [$itk_component(canvas) coords $item] 4 5] break
		set dist [expr sqrt(pow($x-$loc_x,2) + pow($y-$loc_y,2))]
	    } elseif {$tag == "clickable_spotfinding_param"} {
		set l_tags [$itk_component(canvas) gettags $item]
		if {[lsearch $l_tags "beam"] != -1} {
		    foreach {xc_p yc_p} [getCurrentViewCentrePixel "beam"] break
		    set radius_p [expr sqrt(pow($xc_p - $x,2) + pow($yc_p - $y,2))]
		    set radius_mm [expr (double($radius_p) / $zoom) * $pixel_size]
		    set dist $radius_mm
		} elseif {[lsearch $l_tags "backstop_centre"] != -1} {
		    foreach {xc_p yc_p} [getCurrentViewCentrePixel "backstop"] break
		    set radius_p [expr sqrt(pow($xc_p - $x,2) + pow($yc_p - $y,2))]
		    set radius_mm [expr (double($radius_p) / $zoom) * $pixel_size]
		    set dist $radius_mm
		} elseif {[lsearch $l_tags "backstop_radius"] != -1} {
		    foreach {xc_p yc_p} [getCurrentViewCentrePixel "backstop"] break
		    set radius_p [expr sqrt(pow($xc_p - $x,2) + pow($yc_p - $y,2))]
		    set radius_mm [expr (double($radius_p) / $zoom) * $pixel_size]
		    set dist [expr abs($search_area_min_radius - $radius_mm)]
		} elseif {[lsearch $l_tags "search_area_min_radius"] != -1} {
		    foreach {xc_p yc_p} [getCurrentViewCentrePixel "beam"] break
		    set radius_p [expr sqrt(pow($xc_p - $x,2) + pow($yc_p - $y,2))]
		    set radius_mm [expr (double($radius_p) / $zoom) * $pixel_size]
		    set dist [expr abs($search_area_min_radius - $radius_mm)]
		} elseif {[lsearch $l_tags "search_area_max_radius"] != -1} {
		    foreach {xc_p yc_p} [getCurrentViewCentrePixel "beam"] break
		    set radius_p [expr sqrt(pow($xc_p - $x,2) + pow($yc_p - $y,2))]
		    set radius_mm [expr (double($radius_p) / $zoom) * $pixel_size]
		    set dist [expr abs($search_area_max_radius - $radius_mm)]
		} elseif {[lsearch $l_tags vertical_exclusion] != -1} {
		    foreach {xc_p yc_p} [getCurrentViewCentrePixel "beam"] break
		    set width_p [expr abs($xc_p - $x)]
		    set width_mm [expr (double($width_p) / $zoom) * $pixel_size]
		    set dist [expr abs($exclusion_segment_vertical - $width_mm)]
		} elseif {[lsearch $l_tags horizontal_exclusion] != -1} {
		    foreach {xc_p yc_p} [getCurrentViewCentrePixel "beam"] break
		    set width_p [expr abs($yc_p - $y)]
		    set width_mm [expr (double($width_p) / $zoom) * $pixel_size]
		    set dist [expr abs($exclusion_segment_horizontal - $width_mm)]
		} elseif {[lsearch $l_tags high_resolution_limit] != -1} {
		    foreach {xc_p yc_p} [getCurrentViewCentrePixel "beam"] break
		    set radius_p [expr sqrt(pow($xc_p - $x,2) + pow($yc_p - $y,2))]
		    set radius_mm [expr (double($radius_p) / $zoom) * $pixel_size]
		    set dist [expr abs($high_resolution_radius - $radius_mm)]
		} elseif {[lsearch $l_tags low_resolution_limit] != -1} {
		    foreach {xc_p yc_p} [getCurrentViewCentrePixel "beam"] break
		    set radius_p [expr sqrt(pow($xc_p - $x,2) + pow($yc_p - $y,2))]
		    set radius_mm [expr (double($radius_p) / $zoom) * $pixel_size]
		    set dist [expr abs($low_resolution_radius - $radius_mm)]
		} elseif {[lsearch -glob $l_tags mask_polygon(*)] != -1} {
		    set l_poly_tag [lsearch -inline -glob $l_tags mask_polygon(*)]
		    set l_really_overlapped_items [$itk_component(canvas) find overlapping $x $y $x $y]
		    set dist 999
		    foreach i_really_overlapped_item $l_really_overlapped_items {
			set l_really_overlapped_tags [$itk_component(canvas) gettags $i_really_overlapped_item]
			if {[lsearch $l_really_overlapped_tags $l_poly_tag]} {
			    set dist 0
			}
		    }
		} else {
		    set dist 999
		}
	    } else {
		set dist 999
	    }
	    if {$dist < $l_distance} {
		set l_distance $dist
		set chosen $item
	    }
	}
    }
    return $chosen
}



#########################################################
# Pick                                                  #
#########################################################

body ImageDisplay::processPick { a_dom } {
	if {$zoom == $max_zoom} {

    # Clear existing pick
    $itk_component(canvas) delete pick
    # Check on status of task
    set status_code [$a_dom selectNodes string(/pick_region/status/code)]
    if {$status_code == "error"} {
	.m confirm \
	    -type "1button" \
	    -text "[$a_dom selectNodes string(/pick_region/status/message)]" \
	    -button1of2 "Dismiss"

    } else {
	# Extract data
	set l_data [$a_dom selectNodes normalize-space(/pick_region)]	
	# Getcontrast limits
	foreach { l_min l_max } [$itk_component(contrast_palette) getContrast] break
	# get colour threshold
	set l_threshold [expr ($l_min + $l_max) / 2]
	set l_pixel_size $zoom
	set l_y 0
	while {$l_y < 21} {
	    set l_x 0
	    while {$l_x < 21} {
		set l_datum [lindex $l_data [expr ($l_x * 21) + $l_y + 6]]
		if {$l_datum < $l_threshold} {
		    set l_colour black
		} else {
		    set l_colour white
		}
		$itk_component(canvas) create text \
		    [expr ($l_x + 0.5) * $l_pixel_size] \
		    [expr ($l_y + 0.5) * $l_pixel_size] \
		    -text $l_datum \
		    -fill $l_colour \
		    -tags pick
		incr l_x
	    }
	    incr l_y
	}
    }
	} else {
		.pw processPick $a_dom
	}

}

#########################################################
# Methods to be used for displaying bad spots in future #
#########################################################

body ImageDisplay::processBadSpots { a_dom } {
    set badspots_nodes {}
    # Delete existing badspots
    $itk_component(canvas) delete badspots
    $itk_component(mag_canvas) delete badspots

    # Get bad spot information
    set imgno [$a_dom selectNodes normalize-space(/bad_spots_response/image_number)]
    #puts "Processing bad spots for image $imgno"
    set i 0
    foreach bad_spot [$a_dom selectNodes (//bad_spots_response/bad_spot_list/bad_spot)] {
	incr i
	#puts "Bad spot $i: $bad_spot"
	#puts "Bad spot [$bad_spot asHTML]"
	set coordx [$bad_spot selectNodes string(//x_coordinate)]
	set coordy [$bad_spot selectNodes string(//y_coordinate)]
	set units [$bad_spot selectNodes string(//units)]
	set millerh [$bad_spot selectNodes string(//index_h)]
	set millerk [$bad_spot selectNodes string(//index_k)]
	set millerl [$bad_spot selectNodes string(//index_l)]
	set intensity [$bad_spot selectNodes string(//intensity)]
	set sigma [$bad_spot selectNodes string(//sigma)]
	set ratiobgnd [$bad_spot selectNodes string(//background_ratio)]
	set ratiopeak [$bad_spot selectNodes string(//peak_ratio)]
	set bgslopea [$bad_spot selectNodes string(//a_slope)]
	set bgslopeb [$bad_spot selectNodes string(//b_slope)]
	set bgslopec [$bad_spot selectNodes string(//c_slope)]
	set summary [$bad_spot selectNodes string(//summary)]
	#puts "Bad spot $i, $coordx, $coordy, $units, $millerh, $millerk, $millerl, $intensity, $sigma"
	#puts "$ratiobgnd, $ratiopeak, $bgslopea, $bgslopeb, $bgslopec, $summary"
    }

}

body ImageDisplay::plotBadSpots { a_width a_height a_colour a_alt_colour a_coord_list } {
    # Plot markers in pixel positions
    foreach {i_x i_y i_h i_k i_l } $a_coord_list {
	set l_rectangles [plotRectangle $i_x $i_y $a_width $a_height $a_colour $a_alt_colour $i_h $i_k $i_l $itk_component(canvas) $itk_component(mag_canvas)]
    }
}

#########################################################
# Methods used for displaying predicitons               #
#########################################################

body ImageDisplay::getPredictions { } {
    #puts "IMAGE GET PREDICTIONS"
    # Get prediction lists and xml response with width and height from Mosflm
    eval $::mosflm getPredictions
}

body ImageDisplay::clearPredictions { } {
    #puts "IMAGE CLEAR PREDICTIONS"
    $itk_component(canvas) delete prediction
    $itk_component(mag_canvas) delete prediction
    $itk_component(predictions) cancel
    $itk_component(predictions) configure -state "disabled"

    $itk_component(canvas) delete pinpointhkl

}    

body ImageDisplay::findHKL { s_h s_k s_l  } {
    $itk_component(canvas) delete pinpointhkl
    set currenthkl {}
    foreach i_prediction $prediction_nodes {
	foreach {i_x i_y i_h i_k i_l} $i_prediction {
	    if {$i_h == $s_h && $i_k == $s_k && $i_l == $s_l} {
#		$itk_component(canvas) create rectangle  [expr $i_x - 5] [expr $i_y - 5] [expr $i_x + 5] [expr $i_y + 5] -fill red -tags pinpointhkl
		set currenthkl [list $i_x $i_y] 
#		puts "$i_x $i_y $i_h $i_k $i_l"
	    } else {
#		puts "FAILED TO FIND MATCH"
	    }
	}
    }
    plotHKL
}

body ImageDisplay::plotHKL { } {
    $itk_component(canvas) delete pinpointhkl
    foreach {i_x i_y } $currenthkl {
#	puts $currenthkl
#	puts $i_x
#	puts $i_y
#	$itk_component(canvas) create rectangle  [expr $i_x - 5] [expr $i_y - 5] [expr $i_x + 5] [expr $i_y + 5] -fill red -tags pinpointhkl
    $itk_component(canvas) create line [expr $i_x - 15] [expr $i_y - 15] [expr $i_x + 15] [expr $i_y + 15] $i_x $i_y [expr $i_x + 15] [expr $i_y - 15] [expr $i_x - 15] [expr $i_y + 15] -capstyle projecting -fill black -width 2 -tags pinpointhkl
    }
    positionMarkers "pinpointhkl"
}		

body ImageDisplay::processPredictions { a_dom } {
	set prediction_nodes {}
    # Delete existing predictions
    $itk_component(canvas) delete prediction
    $itk_component(mag_canvas) delete prediction

    # Check on status of task
    set status_code [$a_dom selectNodes string(/prediction_response/status/code)]
    if {$status_code == "error"} {
	.m confirm \
	    -type "1button" \
	    -text "[$a_dom selectNodes string(/prediction_response/status/message)]" \
	    -button1of2 "Dismiss"

    } else {
	set l_width [$a_dom selectNodes normalize-space(//width)]
	set l_height [$a_dom selectNodes normalize-space(//height)]
	foreach i_box_list_node [$a_dom selectNodes //boxes/*] {
	    set l_type [$i_box_list_node nodeName]
	    set l_boxes [$i_box_list_node text]
	
		lappend prediction_nodes $l_boxes

	    # Plot each prediction list in their pixel positions
	    if {[llength $l_boxes] > 0} {
		if {$l_type == "fulls"} {
		    set l_colour "\#000000fff"
		    set l_alt_colour "white"
		} elseif  {$l_type == "partials"} {
		    set l_colour "\#ffffff000"
		    set l_alt_colour "black"
		} elseif  {$l_type == "overlaps"} {
		    set l_colour "\#fff000000"
		    set l_alt_colour "black"
		} elseif  {$l_type == "wide"} {
		    set l_colour "\#000fff000"
		    set l_alt_colour "black"
		} else {
		    error "Dodgy prediction type: $l_type"
		}
		plotPredictions $l_width $l_height $l_colour $l_alt_colour $l_boxes
	    } else {
	    }    
	}
    }
    
    # use positionMarkers to relocate them to match the current view
    positionMarkers "prediction"

    # Activate predictions toolbutton if previsouly disabled
    if {[$itk_component(predictions) cget -state] == "disabled"} {
	$itk_component(predictions) configure -state "normal"
	$itk_component(predictions) invoke
    }
    
    # Hide the predictions if the toolbutton was activated
    if {![$itk_component(predictions) query]} {
	$itk_component(canvas) lower prediction view_image
    }

}

body ImageDisplay::plotPredictions { a_width a_height a_colour a_alt_colour a_coord_list } {
    # Plot markers in pixel positions
    foreach {i_x i_y i_h i_k i_l } $a_coord_list {
	set l_rectangles [plotRectangle $i_x $i_y $a_width $a_height $a_colour $a_alt_colour $i_h $i_k $i_l $itk_component(canvas) $itk_component(mag_canvas)]
    }
}


body ImageDisplay::plotRectangle { a_x a_y a_width a_height a_colour a_alt_colour a_h a_k a_l args } {
    
    
    # Caclulate coords of line endings
    set hix [expr $a_x - (double($a_width) / 2) - 1]
    set lox [expr $a_x + (double($a_width) / 2) - 1]
    set hiy [expr $a_y - (double($a_height) / 2) - 1]
    set loy [expr $a_y + (double($a_height) / 2) - 1]
    
    set l_first_canvas 1
    foreach i_canvas $args {
	set l_new_item_1 [$i_canvas create rectangle  $hix $hiy $lox $loy -outline "" -fill "" -tags [list prediction info_colour($a_colour) info_alt_colour($a_alt_colour) marker info_class(prediction) info_label(Prediction:$a_h,$a_k,$a_l)]]
	set l_new_item_2 [$i_canvas create rectangle  $hix $hiy $lox $loy -outline $a_colour -tags [list prediction marker]]
	if {$l_first_canvas} {
	    set l_result [list $l_new_item_1 $l_new_item_2]
	    set l_first_canvas 0
	}
    }
    return $l_result
}

body ImageDisplay::showFindHKLDialog { } {
    if {![winfo exists .fhkl]} {
	FindHKLDialog .fhkl 
    }
	.fhkl setImage [[.image getImage] getShortName]	
	.fhkl confirm
}


class FindHKLDialog {
	inherit Dialog

	private variable search_image
	private variable hvariable ""
	private variable kvariable ""
	private variable lvariable ""
    private method ok
    private method cancel
    public method confirm
	public method setImage

constructor { args } { }
}

body FindHKLDialog::constructor { args } {
#	puts "In the constructor"
    # Matrix combo + label
    itk_component add image_label {
	label $itk_interior.il \
	    -text "Image: "
    }
    itk_component add imagename_label {
	label $itk_interior.inl \
	    -text "[[.image getImage] getShortName]"
    }

    itk_component add h_label {
	label $itk_interior.hl \
	    -text "H:"
    }
    itk_component add k_label {
	label $itk_interior.kl \
	    -text "K:"
    }
    itk_component add l_label {
	label $itk_interior.ll \
	    -text "L:"
    }

    itk_component add h_entry {
	gEntry $itk_interior.hentry \
		-type int \
		-width 7 \
		-justify right \
	    -textvariable [scope hvariable]
    }

    itk_component add k_entry {
	gEntry $itk_interior.kentry \
		-type int \
		-width 7 \
		-justify right \
	    -textvariable [scope kvariable]
    }

    itk_component add l_entry {
	gEntry $itk_interior.lentry \
		-type int \
		-width 7 \
		-justify right \
	    -textvariable [scope lvariable]
    }

    # Buttons

    itk_component add button_frame {
	frame $itk_interior.bf
    }

    itk_component add ok {
	button $itk_interior.bf.ok \
	    -text "Ok" \
	    -width 7 \
	    -command [code $this ok]
    }

    itk_component add cancel {
	button $itk_interior.bf.cancel \
	    -text "Cancel" \
	    -width 7 \
	    -command [code $this cancel]
    }

    grid $itk_component(image_label) $itk_component(imagename_label) -stick w
    grid $itk_component(h_label) $itk_component(h_entry) -stick w
    grid $itk_component(k_label) $itk_component(k_entry) -stick w
    grid $itk_component(l_label) $itk_component(l_entry) -stick w
    grid $itk_component(button_frame) - -sticky we
    pack $itk_component(ok) $itk_component(cancel) \
    -side right \
    -padx {0 7} \
    -pady 7

    eval itk_initialize $args

}


body FindHKLDialog::cancel { } {
    dismiss ""
}

body FindHKLDialog::confirm { } {
    Dialog::confirm
}

body FindHKLDialog::ok { } {
	.image findHKL $hvariable $kvariable $lvariable
#    dismiss ""
}

body FindHKLDialog::setImage {an_image_name} {
	$itk_component(imagename_label) configure -text $an_image_name
}
